'''
See COPYRIGHT.md for copyright information.
'''

from arelle import PluginManager
from arelle.ModelValue import qname
from arelle import XbrlConst
from arelle.Version import copyrightLabel
import regex as re
from collections import defaultdict

def compile(list, traceRows):
    if traceRows:
        # compile so each row can be traced by separate expression (slow)
        return [(rowNbr, re.compile(r"(^|\s)" + pattern + r"($|\W+)", re.IGNORECASE))
                for rowNbr, pattern in list]
    else:
        # compile single expression for fast execution
        return re.compile(r"(^|\s)" +  # always be sure first word starts at start or after space
                          r"($|\W+)|(^|\s)".join(pattern for rowNbr, pattern in list)
                          .replace(r" ",r"\W+") + r"($|\W+)",
                          re.IGNORECASE)

def setup(val, traceRows=False, *args, **kwargs):
    if not val.validateLoggingSemantic:  # all checks herein are SEMANTIC
        return
    # determiniation of two way concept label based on pattern
    # definitions (from documentation label) are used if present, otherwise standard label for these tests
    val.twoWayPriItemDefLabelPattern = compile([
            # from http://www.sec.gov/spotlight/xbrl/staff-review-observations-061511.shtml
            # Cash Flow
            (4, r"increase (\w+ )?decrease"),
            (5, r"provided by (\w+ )?used in"),
            (7, r"net cash inflow or outflow"),
            (6, r"net"),
            (8, r"change in"),
            (9, r"proceeds from (\w+ )?payments (for|to)"),
            # Income statement
            (13, r"(gain|profit) loss"),
            (16, r"income (expense|loss)"),
            (18, r"per share"),
            # Statement of Stockholders Equity
            (22, r"equity"),
            (23, r"retained earnings"),
            # removed? r"conversion of units",
            ], traceRows)
    # standard label tests, indicate two-way label
    val.twoWayPriItemStdLabelPattern = compile([
            # from Eric Cohen
            (4, r"Increase \(Decrease\)"),
            (5, r"Provided by \(Used in\)"),
            (6, r"Net"),
            (8, r"Change in"),
            (9, r"Proceeds from \(Payments for\)"),
            (10, r"Proceeds from \(Payments to\)"),
            (11, r"Payments for \(Proceeds from\)"),
            (12, r"Proceeds from \(Repayments of\)"),
            (13, r"Gain \(Loss\)"),
            (14, r"Profit \(Loss\)"),
            (15, r"Loss \(Gain\)"),
            (16, r"Income \(Loss\)"),
            (17, r"Income \(Expense\)"),
            (18, r"Per Share"),
            (19, r"Per Basic Share"),
            (20, r"Per Diluted Share"),
            (21, r"Per Basic and Diluted"),
            (24, r"Appreciation \(Depreciation\)"),
            (25, r"Asset \(Liability\)"),
            (26, r"Assets Acquired \(Liabilities Assumed\)"),
            (27, r"Benefit \(Expense\)"),
            (28, r"Expense \(Benefit\)"),
            (29, r"Cost[s] \(Credit[s]\)"),
            (30, r"Deductions \(Charges\)"),
            (31, r"Discount \(Premium\)"),
            (32, r"Due from \(to\)"),
            (33, r"Earnings \(Losses\)"),
            (34, r"Earnings \(Deficit\)"),
            (35, r"Excess \(Shortage\)"),
            (36, r"Gains \(Losses\)"),
            (37, r"Impairment \(Recovery\)"),
            (38, r"Income \(Loss\)"),
            (39, r"Liability \(Refund\)"),
            (40, r"Loss \(Recovery\)"),
            (41, r"Obligation[s] \(Asset[s]\)"),
            (42, r"Proceeds from \(Repayments of\)"),
            (43, r"Proceeds from \(Repurchase of\)"),
            (44, r"Provided by \(Used in\)"),
            (45, r"Provisions \(Recoveries\)"),
            (46, r"Retained Earnings \(Accumulated Deficit\)"),
            (47, r"per (\w+ )+"),
            (70, r"Conversion of Units"),
            (71, r"Effective (\w+ )?Rate"),
            ], traceRows)
    # determination of a one-way concept based on standard label
    val.oneWayPriItemDefLabelPattern = compile([
            (49, r"dividend (\w+ )*(paid|received)"),
            ], traceRows)

    val.oneWayPriItemStdLabelPattern = compile([
            (48, r"Payments of (\w+ )*\((Dividends|Capital)\)"),
            (49, r"Dividends (\w+ )*\((Pay(ment)?|Receive|Outstanding)\)"),
            (50, r"(Stock|Shares) Issued"),
            (51, r"Stock (\w+ )*Repurchased"),
            (52, r"(Stock|Shares) (\w+ )*Repurchase[d]?"),
            (53, r"Treasury Stock (\w+ )*(Beginning (\w+ )*Balance[s]?|Ending (\w+ )*Balance[s]?)"),
            (54, r"Treasury Stock (\w+ )*Acquired"),
            (55, r"Treasury Stock (\w+ )*Reissued"),
            (56, r"Treasury Stock (\w+ )*Retired"),
            (57, r"Accumulated Depreciation (\w+ )*Amortization"),
            (58, r"Accumulated Other Than Temporary Impairments"),
            (59, r"Allowance (\w+ )*Doubtful Accounts"),
            (60, r"Amortization (\w+ )*Pension Costs"),
            (61, r"Available for Sale Securities (\w+ )*Continuous Loss Position"),
            (62, r"Available for Sale Securities Bross Unrealized Losses"),
            (63, r"Accounts"),
            ], traceRows)
    # determination of a two way fact based on any of fact's dimension member label
    val.twoWayMemberStdLabelPattern = compile([
            # per Eric Cohen
            (64, r"Change (in|during) \w+"), # don't match word with change in it like exchange
            (65, r"\w+ Elimination \w+"),
            (66, r"Adjustment"),
            (67, r"Effect\s"),
            (68, r"Gain(s)? (\w+ )*Loss(es)?"),
            (69, r"Income \(Loss\)"),
            (70, r"Net(ting)?"),  # don't want to match word with net in it like internet
            ], traceRows)
    val.schedules = {}
    val.elrMatches = (("1statement", re.compile(r"-\s+Statement\s+-\s+", re.IGNORECASE)),
                      ("2disclosure", re.compile(r"-\s+Disclosure\s+-\s+", re.IGNORECASE)),
                      ("3schedule", re.compile(r"-\s+Schedule\s+-\s+", re.IGNORECASE)))

def schedules(val, concept):
    try:
        return val.schedules[concept.qname]
    except KeyError:
        schedules = defaultdict(int)
        for rel in val.modelXbrl.relationshipSet(XbrlConst.parentChild).toModelObject(concept):
            for roleType in val.modelXbrl.roleTypes.get(rel.linkrole,()):
                for elrType, elrPattern in val.elrMatches:
                    if elrPattern.search(roleType.definition):
                        schedules[elrType] += 1
        scheduleStr = ""
        for elrType, num in sorted(schedules.items()):
            scheduleStr += ", {0} {1}{2}".format(num, elrType[1:], "s" if num > 1 else "")
        val.schedules[concept.qname] = scheduleStr
        return scheduleStr


def factCheck(val, fact, *args, **kwargs):
    if not val.validateLoggingSemantic:  # all checks herein are SEMANTIC
        return
    concept = fact.concept
    context = fact.context
    stdLabel = concept.label(lang="en-US", fallbackToQname=False)
    defLabel = concept.label(preferredLabel=XbrlConst.documentationLabel, lang="en-US", fallbackToQname=False)

    try:
        if fact.isNumeric and not fact.isNil and fact.xValue is not None and fact.xValue < 0:
            # is fact an explicit non neg
            if ((defLabel is not None and val.oneWayPriItemDefLabelPattern.search(defLabel)) or
                (stdLabel is not None and val.oneWayPriItemStdLabelPattern.search(stdLabel))):
                if context.qnameDims:  # if fact has a member
                    if any((val.twoWayMemberStdLabelPattern.search(dim.member.label(lang="en-US", fallbackToQname=False))
                            for dim in context.qnameDims.values()
                            if dim.isExplicit)):  # any two way exception member
                        val.modelXbrl.log('INFO-SEMANTIC', "secStaffObservation.nonNegativeFact.info.A",
                            _("Negative fact of an explicit non-negative concept is tagged with a member expected to allow negative values: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                            modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                            value=fact.effectiveValue, elrTypes=schedules(val,concept))
                    else:
                        val.modelXbrl.log('WARNING-SEMANTIC', "secStaffObservation.nonNegativeFact.warning.B",
                            _("Negative fact of an explicit non-negative concept, member may or not justify a negative value: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                            modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                            value=fact.effectiveValue, elrTypes=schedules(val,concept))
                else: # no member
                    val.modelXbrl.log('INCONSISTENCY', "secStaffObservation.nonNegativeFact.inconsistency.C",
                        _("Negative fact of an explicit non-negative concept: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s %(elrTypes)s"),
                        modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                        value=fact.effectiveValue, elrTypes=schedules(val,concept))
            # else test if fact meets two way rules
            elif ((defLabel is not None and val.twoWayPriItemDefLabelPattern.search(defLabel)) or
                  (stdLabel is not None and val.twoWayPriItemStdLabelPattern.search(stdLabel))):
                val.modelXbrl.log('INFO-SEMANTIC', "secStaffObservation.nonNegativeFact.info.D",
                    _("Negative fact of concept expected to have positive and negative values: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                    modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                    value=fact.effectiveValue, elrTypes=schedules(val,concept))
            else:
                if context.qnameDims:  # if fact has a member
                    if any((val.twoWayMemberStdLabelPattern.search(dim.member.label(lang="en-US", fallbackToQname=False))
                            for dim in context.qnameDims.values()
                            if dim.isExplicit)):  # any two way exception member
                        val.modelXbrl.log('INFO-SEMANTIC', "secStaffObservation.nonNegativeFact.info.E",
                            _("Negative fact for typically non-negative concept, but tagged with a member expected to allow negative values: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                            modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                            value=fact.effectiveValue, elrTypes=schedules(val,concept))
                    else:
                        val.modelXbrl.log('WARNING-SEMANTIC', "secStaffObservation.nonNegativeFact.warning.F",
                            _("Negative fact of a typically non-negative concept, member may or not justify a negative value: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                            modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                            value=fact.effectiveValue, elrTypes=schedules(val,concept))
                else: # no member
                    val.modelXbrl.log('INCONSISTENCY', "secStaffObservation.nonNegativeFact.inconsistency.G",
                        _("Negative fact of a \"presumed by default\" non-negative concept: %(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s"),
                        modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
                        value=fact.effectiveValue, elrTypes=schedules(val,concept))
    except Exception as ex:
        val.modelXbrl.log('WARNING-SEMANTIC', "arelle:nonNegFactTestException",
            _("%(fact)s in context %(contextID)s unit %(unitID)s value %(value)s%(elrTypes)s cannot be tested nonnegative"),
            modelObject=fact, fact=fact.qname, contextID=fact.contextID, unitID=fact.unitID,
            value=fact.effectiveValue, elrTypes=schedules(val,fact))

def final(val, conceptsUsed, *args, **kwargs):
    if not val.validateLoggingSemantic:  # all checks herein are SEMANTIC
        return
    del val.twoWayPriItemDefLabelPattern
    del val.twoWayPriItemStdLabelPattern
    del val.oneWayPriItemStdLabelPattern
    del val.twoWayMemberStdLabelPattern
    del val.schedules


def saveDtsMatches(dts, secDtsTagMatchesFile):
    setup(dts, True)
    import sys, csv
    csvOpenMode = 'w'
    csvOpenNewline = ''
    csvFile = open(secDtsTagMatchesFile, csvOpenMode, newline=csvOpenNewline)
    csvWriter = csv.writer(csvFile, dialect="excel")
    csvWriter.writerow(("Concept", "Rule", "Row", "Pattern", "Label", "Documentation"))
    num1wayConcepts = 0
    num2wayConcepts = 0
    num2wayMembers = 0

    for qname, concept in sorted(dts.qnameConcepts.items(), key=lambda item: item[0]):
        if concept.isItem and concept.isPrimaryItem: # both pri item and domain members
            stdLabel = concept.label(lang="en-US", fallbackToQname=False)
            defLabel = concept.label(preferredLabel=XbrlConst.documentationLabel, lang="en-US", fallbackToQname=False)
            if concept.type is not None and concept.type.isDomainItemType:
                if stdLabel is not None:
                    for rowNbr, pattern in dts.twoWayMemberStdLabelPattern:
                        if pattern.search(stdLabel):
                            csvWriter.writerow((str(qname), "member-2-way", rowNbr, pattern.pattern[6:-7], stdLabel, defLabel))
                            num2wayMembers += 1
            elif concept.isNumeric and not concept.isAbstract: # not dimension domain/member
                if defLabel is not None:
                    for rowNbr, pattern in dts.twoWayPriItemDefLabelPattern:
                        if pattern.search(defLabel):
                            csvWriter.writerow((str(qname), "concept-2-way-doc", rowNbr, pattern.pattern[6:-7], stdLabel, defLabel))
                            num2wayConcepts += 1
                    for rowNbr, pattern in dts.oneWayPriItemDefLabelPattern:
                        if pattern.search(defLabel):
                            csvWriter.writerow((str(qname), "concept-1-way-doc", rowNbr, pattern.pattern[6:-7], stdLabel, defLabel))
                            num1wayConcepts += 1
                if stdLabel is not None:
                    for rowNbr, pattern in dts.twoWayPriItemStdLabelPattern:
                        if pattern.search(stdLabel):
                            csvWriter.writerow((str(qname), "concept-2-way-lbl", rowNbr, pattern.pattern[6:-7], stdLabel, defLabel))
                            num2wayConcepts += 1
                    for rowNbr, pattern in dts.oneWayPriItemStdLabelPattern:
                        if pattern.search(stdLabel):
                            csvWriter.writerow((str(qname), "concept-1-way-lbl", rowNbr, pattern.pattern[6:-7], stdLabel, defLabel))
                            num1wayConcepts += 1

    csvFile.close()

    dts.log('INFO-SEMANTIC', "info:saveSecDtsTagMatches",
             _("SecDtsTagMatches entry %(entryFile)s has %(numberOfTwoWayPriItems)s two way primary items, %(numberOfOneWayPriItems)s one way primary items, %(numberOfTwoWayMembers)s two way members in output file %(secDtsTagMatchesFile)s."),
             modelObject=dts,
             entryFile=dts.uri,
             numberOfTwoWayPriItems=num2wayConcepts,
             numberOfOneWayPriItems=num1wayConcepts,
             numberOfTwoWayMembers=num2wayMembers,
             secDtsTagMatchesFile=secDtsTagMatchesFile)

    final(dts)

def saveDtsMatchesMenuEntender(cntlr, menu, *args, **kwargs):
    # Extend menu with an item for the savedts plugin
    menu.add_command(label="Save SEC tag matches",
                     underline=0,
                     command=lambda: saveDtsMatchesMenuCommand(cntlr) )

def saveDtsMatchesMenuCommand(cntlr):
    # save DTS menu item has been invoked
    if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
        cntlr.addToLog("No taxonomy loaded.")
        return

        # get file name into which to save log file while in foreground thread
    secDtsTagMatchesFile = cntlr.uiFileDialog("save",
            title=_("Save SEC DTS tag matches file"),
            filetypes=[(_("DTS tag matches .csv file"), "*.csv")],
            defaultextension=".txt")
    if not secDtsTagMatchesFile:
        return False

    try:
        saveDtsMatches(cntlr.modelManager.modelXbrl, secDtsTagMatchesFile)
    except Exception as ex:
        dts = cntlr.modelManager.modelXbrl
        dts.error("exception",
            _("SEC DTS Tags Matches exception: %(error)s"), error=ex,
            modelXbrl=dts,
            exc_info=True)

def saveDtsMatchesCommandLineOptionExtender(parser, *args, **kwargs):
    # extend command line options with a save DTS option
    parser.add_option("--save-sec-tag-dts-matches",
                      action="store",
                      dest="secDtsTagMatchesFile",
                      help=_("Save SEC DTS tag matches CSV file."))

def saveDtsMatchesCommandLineXbrlRun(cntlr, options, modelXbrl, *args, **kwargs):
    # extend XBRL-loaded run processing for this option
    if getattr(options, "secDtsTagMatchesFile", False):
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No taxonomy loaded.")
            return
        saveDtsMatches(cntlr.modelManager.modelXbrl, options.secDtsTagMatchesFile)

__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Validate US SEC Tagging',
    'version': '0.9',
    'description': '''US SEC Tagging Validation.  Includes non-negative rules.''',
    'license': 'Apache-2',
    'author': 'Ewe S. Gap',
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'Validate.EFM.Start': setup,
    'Validate.EFM.Fact': factCheck,
    'Validate.EFM.Finally': final,
    'CntlrWinMain.Menu.Tools': saveDtsMatchesMenuEntender,
    'CntlrCmdLine.Options': saveDtsMatchesCommandLineOptionExtender,
    'CntlrCmdLine.Xbrl.Run': saveDtsMatchesCommandLineXbrlRun,
}



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations
from arelle.Version import __version__
from enum import Enum, auto
from typing import Any
import math
import os
import platform
import sys


class PlatformOS(Enum):
    LINUX = auto()
    MACOS = auto()
    WINDOWS = auto()

    @staticmethod
    def getPlatformOS() -> PlatformOS:
        """
        Using system.platform() raises a value error exception
        in windows testing envs and macOS with Python 3.8
        """
        if sys.platform == "darwin":
            return PlatformOS.MACOS
        if sys.platform.startswith("win"):
            return PlatformOS.WINDOWS
        # For legacy support purposes all other platforms are treated the same as Linux.
        return PlatformOS.LINUX


def isCGI() -> bool:
    return os.getenv("GATEWAY_INTERFACE", "").startswith("CGI/")


def isGAE() -> bool:
    serverSoftware = os.getenv("SERVER_SOFTWARE", "")
    return serverSoftware.startswith("Google App Engine/") or serverSoftware.startswith("Development/")


def hasFileSystem() -> bool:
    return not isGAE()


def hasWebServer() -> bool:
    try:
        from arelle import webserver
        return True
    except ImportError:
        return False


def getSystemWordSize() -> int:
    return int(round(math.log(sys.maxsize, 2)) + 1)


def hasVirtualEnv() -> bool:
    return getattr(sys, "base_prefix", sys.prefix) != sys.prefix or hasattr(sys, "real_prefix")


def getSystemInfo() -> dict[str, Any]:
    """Return info about the system."""
    info_object = {
        "arelle_version": __version__,
        "arch": platform.machine(),
        "args": sys.argv,
        "cgi": isCGI(),
        "filesystem": hasFileSystem(),
        "gae": isGAE(),
        "docker": False,
        "os_name": platform.system(),
        "os_version": platform.release(),
        "python_branch": platform.python_branch(),
        "python_compiler": platform.python_compiler(),
        "python_implementation": platform.python_implementation(),
        "python_version": platform.python_version(),
        "python_virtualenv": hasVirtualEnv(),
        "system_word_size": getSystemWordSize(),
        "webserver": hasWebServer(),
    }

    if PlatformOS.getPlatformOS() == PlatformOS.MACOS:
        info_object["os_version"] = platform.mac_ver()[0]
    elif PlatformOS.getPlatformOS() == PlatformOS.LINUX:
        info_object["docker"] = os.path.isfile("/.dockerenv")
    return info_object



import os
from pathlib import PurePath, Path
from tests.integration_tests.validation.conformance_suite_config import (
    CONFORMANCE_SUITE_PATH_PREFIX, ConformanceSuiteConfig, ConformanceSuiteAssetConfig, AssetSource
)

ZIP_PATH = Path('inlineXBRL-1.1-conformanceSuite-2020-04-08.zip')
# needs to be extracted because arelle can't load a taxonomy package ZIP from within a ZIP
EXTRACTED_PATH = Path(ZIP_PATH.stem)
config = ConformanceSuiteConfig(
    args=[
        '--packages', os.path.join(CONFORMANCE_SUITE_PATH_PREFIX, EXTRACTED_PATH, 'inlineXBRL-1.1-conformanceSuite-2020-04-08/schemas/www.example.com.zip'),
    ],
    assets=[
        ConformanceSuiteAssetConfig.nested_conformance_suite(
            ZIP_PATH,
            EXTRACTED_PATH,
            entry_point_root=EXTRACTED_PATH,
            entry_point=Path('inlineXBRL-1.1-conformanceSuite-2020-04-08/index.xml'),
            public_download_url='https://www.xbrl.org/2020/inlineXBRL-1.1-conformanceSuite-2020-04-08.zip',
            source=AssetSource.S3_PUBLIC,
        ),
    ],
    capture_warnings=False,
    info_url='https://specifications.xbrl.org/work-product-index-inline-xbrl-inline-xbrl-1.1.html',
    name=PurePath(__file__).stem,
    network_or_cache_required=False,
    plugins=frozenset({'inlineXbrlDocumentSet', '../../tests/plugin/testcaseIxExpectedHtmlFixup.py'}),
    shards=4,
)



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass

from arelle.ModelInstanceObject import ModelUnit, ModelContext, ModelFact
from arelle.ModelValue import QName
from arelle.ModelXbrl import ModelXbrl
from arelle.utils.PluginData import PluginData


@dataclass
class PluginValidationDataExtension(PluginData):
    financialReportingPeriodCurrentStartDateQn: QName
    financialReportingPeriodCurrentEndDateQn: QName
    financialReportingPeriodPreviousStartDateQn: QName
    financialReportingPeriodPreviousEndDateQn: QName
    formattedExplanationItemTypeQn: QName
    documentAdoptionDateQn: QName
    documentAdoptionStatusQn: QName
    documentResubmissionUnsurmountableInaccuraciesQn: QName
    entrypointRoot: str
    entrypoints: set[str]
    textFormattingSchemaPath: str
    textFormattingWrapper: str

    _contextsByDocument: dict[str, list[ModelContext]] | None = None
    _factsByDocument: dict[str, list[ModelFact]] | None = None
    _unitsByDocument: dict[str, list[ModelUnit]] | None = None

    def contextsByDocument(self, modelXbrl: ModelXbrl) -> dict[str, list[ModelContext]]:
        if self._contextsByDocument is not None:
            return self._contextsByDocument
        contextsByDocument = defaultdict(list)
        for context in modelXbrl.contexts.values():
            contextsByDocument[context.modelDocument.filepath].append(context)
        self._contextsByDocument = dict(contextsByDocument)
        return self._contextsByDocument

    def factsByDocument(self, modelXbrl: ModelXbrl) -> dict[str, list[ModelFact]]:
        if self._factsByDocument is not None:
            return self._factsByDocument
        factsByDocument = defaultdict(list)
        for fact in modelXbrl.facts:
            factsByDocument[fact.modelDocument.filepath].append(fact)
        self._factsByDocument = dict(factsByDocument)
        return self._factsByDocument

    def unitsByDocument(self, modelXbrl: ModelXbrl) -> dict[str, list[ModelUnit]]:
        if self._unitsByDocument is not None:
            return self._unitsByDocument
        unitsByDocument = defaultdict(list)
        for unit in modelXbrl.units.values():
            unitsByDocument[unit.modelDocument.filepath].append(unit)
        self._unitsByDocument = dict(unitsByDocument)
        return self._unitsByDocument



'''
See COPYRIGHT.md for copyright information.
'''
import os
from arelle import XmlUtil
from arelle.ModelDocument import ModelDocument, Type

class ModelRssObject(ModelDocument):
    """
    .. class:: ModelRssObject(type=ModelDocument.Type.RSSFEED, uri=None, filepath=None, xmlDocument=None)

    ModelRssObject is a specialization of ModelDocument for RSS Feeds.

    (for parameters and inherited attributes, please see ModelDocument)
    """
    def __init__(self, modelXbrl,
                 type=Type.RSSFEED,
                 uri=None, filepath=None, xmlDocument=None):
        super(ModelRssObject, self).__init__(modelXbrl, type, uri, filepath, xmlDocument)
        self.rssItems = []

    def rssFeedDiscover(self, rootElement):
        """Initiates discovery of RSS feed
        """
        # add self to namespaced document
        self.xmlRootElement = rootElement
        for itemElt in XmlUtil.descendants(rootElement, None, "item"):
            self.rssItems.append(itemElt)



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

import unicodedata
from collections import defaultdict

import regex as re

from arelle import ModelDocument as ModelDocumentFile, XbrlConst
from arelle.ModelDocument import ModelDocument
from arelle.ModelDtsObject import ModelConcept, ModelType
from arelle.ModelObject import ModelObject
from arelle.ValidateXbrl import ValidateXbrl
from arelle.XbrlConst import dimensionDefault, standardLabelRoles, xbrli
from arelle.typing import TypeGetText
from ..Const import (
    DefaultDimensionLinkroles,
    esefDefinitionArcroles,
    filenamePatterns,
    filenameRegexes,
    linkbaseRefTypes,
    qnDomainItemTypes,
    qnDomainItemTypes2023,
)
from ..Util import isChildOfNotes, isExtension, getDisclosureSystemYear

_: TypeGetText  # Handle gettext


def checkFilingDTS(val: ValidateXbrl, modelDocument: ModelDocument, esefNotesConcepts: set[str],
                   visited: list[ModelDocument], ifrsNses: list[str], hrefXlinkRole: str | None =None) -> None:
    visited.append(modelDocument)
    for referencedDocument, modelDocumentReference in modelDocument.referencesDocument.items():
        if referencedDocument not in visited and referencedDocument.inDTS: # ignore non-DTS documents
            checkFilingDTS(val, referencedDocument, esefNotesConcepts,
                           visited, ifrsNses, modelDocumentReference.referringXlinkRole)

    isExtensionDoc = isExtension(val, modelDocument)
    filenamePattern = filenameRegex = None
    anchorAbstractExtensionElements = getDisclosureSystemYear(val.modelXbrl) < 2023 and val.authParam["extensionElementsAnchoring"] == "include abstract"
    allowCapsInLc3Words = val.authParam["LC3AllowCapitalsInWord"]

    def lc3wordAdjust(word: str) -> str:
        if allowCapsInLc3Words:
            return word.title()
        elif len(word) > 1:
            return word[0].upper() + word[1:]
        return word

    if not isExtensionDoc:
        pass

    # the following doc type sections only pertain to extensionDocuments
    elif modelDocument.type == ModelDocumentFile.Type.INLINEXBRL:
        if val.authParam["reportFileNamePattern"]:
            filenamePattern = val.authParam["reportFileNamePattern"]
            filenameRegex = val.authParam["reportFileNameRegex"]

    elif modelDocument.type == ModelDocumentFile.Type.SCHEMA:

        val.hasExtensionSchema = True

        filenamePattern = "{base}-{date}.xsd"
        filenameRegex = r"(.{1,})-[0-9]{4}-[0-9]{2}-[0-9]{2}[.]xsd$"

        for doc, docRef in modelDocument.referencesDocument.items():
            if "import" in docRef.referenceTypes:
                val.extensionImportedUrls.add(doc.uri)
            #if docRef.referenceTypes & {"import","include"}:
            #    if disallowedURIsPattern.match(doc.uri):
            #        val.modelXbrl.warning("ESEF.3.1.1.extensionImportNotAllowed",
            #                            _("Taxonomy reference not allowed for extension schema: %(taxonomy)s"),
            #                            modelObject=modelDocument, taxonomy=doc.uri)

        tuplesInExtTxmy = []
        fractionsInExtTxmy = []
        typedDimsInExtTxmy = []
        domainMembersWrongType = []
        extLineItemsWithoutHypercube = []
        extLineItemsNotAnchored = []
        extLineItemsWronglyAnchored = []
        extAbstractConcepts = []
        extMonetaryConceptsWithoutBalance = []
        conceptsWithoutStandardLabel = []
        conceptsWithNoLabel = []
        parentChildRelSet = val.modelXbrl.relationshipSet(XbrlConst.parentChild)
        widerNarrowerRelSet = val.modelXbrl.relationshipSet(XbrlConst.widerNarrower)
        generalSpecialRelSet = val.modelXbrl.relationshipSet(XbrlConst.generalSpecial)
        calcRelSet = val.modelXbrl.relationshipSet(XbrlConst.summationItem)
        dimensionDefaults = val.modelXbrl.relationshipSet(dimensionDefault, DefaultDimensionLinkroles)
        labelsRelationshipSet = val.modelXbrl.relationshipSet(XbrlConst.conceptLabel)
        if modelDocument.targetNamespace is not None:
            for modelConcept in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.w3.org/2001/XMLSchema}element"):
                if isinstance(modelConcept,ModelConcept):
                    name = modelConcept.get("name")
                    if name is None:
                        name = ""
                        if modelConcept.get("ref") is not None:
                            continue    # don't validate ref's here
                    if modelConcept.isTuple:
                        tuplesInExtTxmy.append(modelConcept)
                    if modelConcept.isFraction:
                        fractionsInExtTxmy.append(modelConcept)
                    if modelConcept.isTypedDimension:
                        typedDimsInExtTxmy.append(modelConcept)
                    if modelConcept.isExplicitDimension and not dimensionDefaults.fromModelObject(modelConcept):
                        val.modelXbrl.error("ESEF.3.4.3.extensionTaxonomyDimensionNotAssignedDefaultMemberInDedicatedPlaceholder",
                            _("Each dimension in an issuer specific extension taxonomy MUST be assigned to a default member in the ELR with role URI http://www.esma.europa.eu/xbrl/role/core/ifrs-dim_role-990000 defined in esef_cor.xsd schema file. %(qname)s"),
                            modelObject=modelConcept, qname=modelConcept.qname)
                    esefDomainItemTypes = qnDomainItemTypes if getDisclosureSystemYear(val.modelXbrl) < 2023 else qnDomainItemTypes2023
                    if modelConcept.isDomainMember and modelConcept in val.domainMembers and modelConcept.typeQname not in esefDomainItemTypes:
                        domainMembersWrongType.append(modelConcept)
                    if modelConcept.isPrimaryItem and not modelConcept.isAbstract:
                        if modelConcept not in val.primaryItems:
                            extLineItemsWithoutHypercube.append(modelConcept)
                        elif not widerNarrowerRelSet.fromModelObject(modelConcept) and not widerNarrowerRelSet.toModelObject(modelConcept):
                            # Reporting manual - 1.4 Anchoring -> RTS on ESEF does not set an anchoring requirement for the Notes
                            # to the financial statements
                            conceptRels = parentChildRelSet.toModelObject(modelConcept)
                            conceptLinkroles = tuple(set(rel.linkrole for rel in conceptRels))
                            conceptLinkroleRestrictedRelSet = val.modelXbrl.relationshipSet(XbrlConst.parentChild,
                                                                                            conceptLinkroles)
                            # Globally, this has O(extensions*presentation) running time,
                            # which could be slow if there are many unanchored extensions.
                            # This could be improved by precomputing childrenOfNotes
                            if not calcRelSet.fromModelObject(modelConcept) and not isChildOfNotes(modelConcept,
                                                                                                   conceptLinkroleRestrictedRelSet,
                                                                                                   esefNotesConcepts,
                                                                                                   set()): # exclude subtotals
                                # Conformance suite RTS_Annex_IV_Par_9_Par_10_G1-4-1_G1-4-2_G3-3-1_G3-3-2/TC6_invalid: look for other arcroles
                                if not generalSpecialRelSet.fromModelObject(modelConcept) and not generalSpecialRelSet.toModelObject(modelConcept):
                                    extLineItemsNotAnchored.append(modelConcept)
                                else:
                                    extLineItemsWronglyAnchored.append(modelConcept)
                    if (modelConcept.isAbstract and modelConcept not in val.domainMembers and
                        modelConcept.type is not None and not modelConcept.type.isDomainItemType and
                        not modelConcept.isHypercubeItem and not modelConcept.isDimensionItem):
                        extAbstractConcepts.append(modelConcept)
                        if anchorAbstractExtensionElements:
                            if not widerNarrowerRelSet.fromModelObject(modelConcept) and not widerNarrowerRelSet.toModelObject(modelConcept):
                                # Conformance suite RTS_Annex_IV_Par_9_Par_10_G1-4-1_G1-4-2_G3-3-1_G3-3-2/TC6_invalid: look for other arcroles
                                if not generalSpecialRelSet.fromModelObject(modelConcept) and not generalSpecialRelSet.toModelObject(modelConcept):
                                    extLineItemsNotAnchored.append(modelConcept)
                                else:
                                    extLineItemsWronglyAnchored.append(modelConcept)
                    if modelConcept.isMonetary and not modelConcept.balance:
                        extMonetaryConceptsWithoutBalance.append(modelConcept)
                    # check all lang's of standard label
                    hasLc3Match = False
                    lc3names = []
                    hasStandardLabel = False
                    hasNonStandardLabel = False
                    if labelsRelationshipSet:
                        for labelRel in labelsRelationshipSet.fromModelObject(modelConcept):
                            concept = labelRel.fromModelObject
                            label = labelRel.toModelObject
                            if concept is not None and label is not None:
                                if label.role == XbrlConst.standardLabel:
                                    hasStandardLabel = True
                                    # allow Joe's Bar, N.A.  to be JoesBarNA -- remove ', allow A. as not article "a"
                                    lc3name = ''.join(re.sub(r"['.-]", "", lc3wordAdjust(w[0] or w[2] or w[3] or w[4]))
                                                      for w in re.findall(r"((\w+')+\w+)|(A[.-])|([.-]A(?=\W|$))|(\w+)",
                                                                          unicodedata.normalize('NFKD', label.textValue)
                                                                          .encode('ASCII', 'ignore').decode()  # remove diacritics
                                                                          )
                                                      # if w[4].lower() not in ("the", "a", "an")
                                                      )
                                    lc3names.append(lc3name)
                                    if (name == lc3name or
                                        (name and lc3name and lc3name[0].isdigit() and name[1:] == lc3name and (name[0].isalpha() or name[0] == '_'))):
                                        hasLc3Match = True
                                else:
                                    hasNonStandardLabel = True
                            if label.role not in standardLabelRoles and not ( #not in LRR
                                label.role in val.modelXbrl.roleTypes and val.modelXbrl.roleTypes[label.role][0].modelDocument.uri.startswith("http://www.xbrl.org/lrr")):
                                val.modelXbrl.warning("ESEF.3.4.5.taxonomyElementLabelCustomRole",
                                    _("Extension taxonomy element label SHOULD not be custom: %(concept)s role %(labelrole)s"),
                                    modelObject=(modelConcept,label), concept=modelConcept.qname, labelrole=label.role)
                    if modelConcept.isItem or modelConcept.isTuple or modelConcept.isHypercubeItem or modelConcept.isDimensionItem:
                        if not hasStandardLabel:
                            if hasNonStandardLabel:
                                conceptsWithoutStandardLabel.append(modelConcept)
                            else:
                                conceptsWithNoLabel.append(modelConcept)
            for modelType in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.w3.org/2001/XMLSchema}complexType"):
                if (isinstance(modelType,ModelType) and isExtension(val, modelType) and
                    modelType.typeDerivedFrom is not None and modelType.typeDerivedFrom.qname.namespaceURI == xbrli and
                    not modelType.particlesList):
                    val.modelXbrl.error("ESEF.RTS.Annex.IV.Par.11.customDataTypeDuplicatingXbrlOrDtrEntry",
                        _("Extension taxonomy element must not define a type where one is already defined by the XBRL specifications or in the XBRL Data Types Registry: %(qname)s"),
                        modelObject=modelType, qname=modelType.qname)
        if tuplesInExtTxmy:
            val.modelXbrl.error("ESEF.2.4.1.tupleDefinedInExtensionTaxonomy",
                _("Tuples MUST NOT be defined in extension taxonomy: %(concepts)s"),
                modelObject=tuplesInExtTxmy, concepts=", ".join(str(c.qname) for c in tuplesInExtTxmy))
        if fractionsInExtTxmy:
            val.modelXbrl.error("ESEF.2.4.1.fractionDefinedInExtensionTaxonomy",
                _("Fractions MUST NOT be defined in extension taxonomy: %(concepts)s"),
                modelObject=fractionsInExtTxmy, concepts=", ".join(str(c.qname) for c in fractionsInExtTxmy))
        if typedDimsInExtTxmy:
            val.modelXbrl.error("ESEF.3.2.3.typedDimensionDefinitionInExtensionTaxonomy",
                _("Extension taxonomy MUST NOT define typed dimensions: %(concepts)s."),
                modelObject=typedDimsInExtTxmy, concepts=", ".join(str(c.qname) for c in typedDimsInExtTxmy))
        if domainMembersWrongType:
            xbrlReference322 = "https://www.xbrl.org/dtr/type/2020-01-21/types.xsd"
            if getDisclosureSystemYear(val.modelXbrl) < 2023:
                xbrlReference322 = "http://www.xbrl.org/dtr/type/nonNumeric-2009-12-16.xsd"
            val.modelXbrl.error("ESEF.3.2.2.domainMemberWrongDataType",
                _("Domain members MUST have domainItemType data type as defined in \"%(xbrlReference)s\": concept %(concepts)s."),
                modelObject=domainMembersWrongType, xbrlReference=xbrlReference322,
                concepts=", ".join(str(c.qname) for c in domainMembersWrongType))
        # HF - think this is only about reported line items, not ext line items (?)
        #if extLineItemsWithoutHypercube:
        #    val.modelXbrl.error("ESEF.3.4.2.extensionTaxonomyLineItemNotLinkedToAnyHypercube",
        #        _("Line items that do not require any dimensional information to tag data MUST be linked to \"Line items not dimensionally qualified\" hypercube in http://www.esma.europa.eu/xbrl/esef/role/esef_role-999999 declared in esef_cor.xsd: concept %(concepts)s."),
        #        modelObject=extLineItemsWithoutHypercube, concepts=", ".join(str(c.qname) for c in extLineItemsWithoutHypercube))
        if extLineItemsNotAnchored:
            val.modelXbrl.error("ESEF.3.3.1.extensionConceptsNotAnchored",
                _("Extension concepts SHALL be anchored to concepts in the ESEF taxonomy:  %(concepts)s."),
                modelObject=extLineItemsNotAnchored, concepts=", ".join(str(c.qname) for c in extLineItemsNotAnchored))
        if extLineItemsWronglyAnchored:
            val.modelXbrl.error("ESEF.3.3.1.anchoringWrongArcrole",
                _("Anchoring relationships for concepts MUST use "
                  "\"http://www.esma.europa.eu/xbrl/esef/arcrole/wider-narrower\" arcrole: %(concepts)s."),
                modelObject=extLineItemsWronglyAnchored, concepts=", ".join(sorted(str(c.qname) for c in extLineItemsWronglyAnchored)))
        if extMonetaryConceptsWithoutBalance:
            val.modelXbrl.error("ESEF.RTS.Annex.IV.Par.4.2.monetaryConceptWithoutBalance",
                _("Extension monetary concepts MUST provide balance attribute: concept %(concepts)s."),
                modelObject=extMonetaryConceptsWithoutBalance, concepts=", ".join(str(c.qname) for c in extMonetaryConceptsWithoutBalance))
        if conceptsWithNoLabel:
            val.modelXbrl.error("ESEF.RTS.Annex.IV.Par.4_G3.4.5.extensionConceptNoLabel",
                _("Extension concepts MUST provide labels: %(concepts)s."),
                modelObject=conceptsWithNoLabel, concepts=", ".join(str(c.qname) for c in conceptsWithNoLabel))
        if conceptsWithoutStandardLabel:
            val.modelXbrl.error("ESEF.RTS.Annex.IV.Par.4_G3.4.5.extensionConceptNoStandardLabel",
                _("Extension concepts MUST provide a standard label: %(concepts)s."),
                modelObject=conceptsWithoutStandardLabel, concepts=", ".join(str(c.qname) for c in conceptsWithoutStandardLabel))

        embeddedLinkbaseElements = [e
                                    for e in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.xbrl.org/2003/linkbase}linkbase")
                                    if isinstance(e,ModelObject)]
        if embeddedLinkbaseElements:
                val.modelXbrl.error("ESEF.3.1.1.linkbasesNotSeparateFiles",
                    _("Each linkbase type MUST be provided in a separate linkbase file, but a linkbase was found in %(schema)s."),
                    modelObject=embeddedLinkbaseElements, schema=modelDocument.basename)

        del (tuplesInExtTxmy, fractionsInExtTxmy, typedDimsInExtTxmy, domainMembersWrongType, generalSpecialRelSet,
             extLineItemsWithoutHypercube, extLineItemsNotAnchored, extLineItemsWronglyAnchored, extAbstractConcepts,
             extMonetaryConceptsWithoutBalance, conceptsWithNoLabel, conceptsWithoutStandardLabel)

    elif modelDocument.type == ModelDocumentFile.Type.LINKBASE:

        linkbasesFound = set()
        disallowedArcroles = defaultdict(list)
        prohibitedBaseConcepts = []
        prohibitingLbElts = []
        linkbaseRefType = None

        if hrefXlinkRole in linkbaseRefTypes:
            linkbaseRefType = linkbaseRefTypes[hrefXlinkRole]
            filenamePattern = filenamePatterns[linkbaseRefType]
            filenameRegex = filenameRegexes[linkbaseRefType]

        linkEltName: str | None # linkEltName is set to None further down
        for linkEltName in ("labelLink", "presentationLink", "calculationLink", "definitionLink", "referenceLink"):
            for linkElt in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.xbrl.org/2003/linkbase}" + linkEltName):
                if not filenamePattern:
                    filenamePattern = filenamePatterns.get(linkEltName[:3])
                    filenameRegex = filenameRegexes.get(linkEltName[:3])
                if linkEltName == "labelLink":
                    val.hasExtensionLbl = True
                    linkbasesFound.add(linkEltName)
                if linkEltName == "presentationLink":
                    val.hasExtensionPre = True
                    linkbasesFound.add(linkEltName)
                if linkEltName == "calculationLink":
                    val.hasExtensionCal = True
                    linkbasesFound.add(linkEltName)
                if linkEltName == "definitionLink":
                    val.hasExtensionDef = True
                    linkbasesFound.add(linkEltName)
                    if not filenamePattern:
                        filenamePattern = filenamePatterns["lab"]
                        filenameRegex = filenameRegexes["lab"]
                    # check for any unexpected definition arcrole which might be a custom wider-narrower arcrole
                    for arcElt in linkElt.iterchildren(tag="{http://www.xbrl.org/2003/linkbase}definitionArc"):
                        arcrole = arcElt.get("{http://www.w3.org/1999/xlink}arcrole")
                        if arcrole not in esefDefinitionArcroles:
                            disallowedArcroles[arcrole].append(arcElt)

                if linkEltName in ("definitionLink", ) and getDisclosureSystemYear(val.modelXbrl) == 2023 and val.authParam["validate1_9_1"] in ("true", "True", 1):
                    for locElt in linkElt.iterchildren("{http://www.xbrl.org/2003/linkbase}loc"):
                        refObject = locElt.dereference()
                        if (isinstance(refObject, ModelConcept)
                                and refObject.qname.namespaceURI in ifrsNses
                                and refObject.qname.localName == "DisclosureOfNotesAndOtherExplanatoryInformationExplanatory"):
                            val.modelXbrl.warning(
                                "ESEF.1.9.1.disclosureOfNotesAndOtherExplanatoryInformationExplanatoryDeprecated",
                                _("Usage of concept 'DisclosureOfNotesAndOtherExplanatoryInformationExplanatory' is deprecated."),
                                modelObject=locElt)
                if linkEltName in ("labelLink", "referenceLink"):
                    # check for prohibited esef taxnonomy target elements
                    prohibitedArcFroms = defaultdict(list)
                    prohibitedArcTos = defaultdict(list)
                    for arcElt in linkElt.iterchildren("{http://www.xbrl.org/2003/linkbase}labelArc", "{http://www.xbrl.org/2003/linkbase}referenceArc"):
                        if arcElt.get("use") == "prohibited":
                             prohibitedArcFroms[arcElt.get("{http://www.w3.org/1999/xlink}from")].append(arcElt)
                             prohibitedArcTos[arcElt.get("{http://www.w3.org/1999/xlink}to")].append(arcElt)
                    for locElt in linkElt.iterchildren("{http://www.xbrl.org/2003/linkbase}loc"):
                        prohibitingArcs = prohibitedArcTos.get(locElt.get("{http://www.w3.org/1999/xlink}label"))
                        if prohibitingArcs and not isExtension(val, locElt.get("{http://www.w3.org/1999/xlink}href")):
                            prohibitingLbElts.extend(prohibitingArcs)
                        prohibitingArcs = prohibitedArcFroms.get(locElt.get("{http://www.w3.org/1999/xlink}label"))
                        if prohibitingArcs and not isExtension(val, locElt.get("{http://www.w3.org/1999/xlink}href")):
                            prohibitingLbElts.extend(prohibitingArcs)
                            prohibitedBaseConcepts.append(locElt.dereference())
                    del prohibitedArcFroms, prohibitedArcTos # dereference
            for disallowedArcrole, arcs in disallowedArcroles.items():
                val.modelXbrl.error("ESEF.RTS.Annex.IV.disallowedDefinitionArcrole",
                    _("Disallowed arcrole in definition linkbase %(arcrole)s."),
                    modelObject=arcs, arcrole=arcrole)
            disallowedArcroles.clear()
            if prohibitingLbElts and prohibitedBaseConcepts:
                val.modelXbrl.error("ESEF.RTS.Annex.IV.Par.8.coreTaxonomy{}Modification".format(linkEltName[:-4].title()),
                    _("Disallowed modification of core taxonomy %(resource)s for %(qnames)s."),
                    modelObject=prohibitingLbElts+prohibitedBaseConcepts,
                    resource=linkEltName[:-4],
                    qnames=", ".join(str(c.qname) for c in prohibitedBaseConcepts),
                    messageCodes=("ESEF.RTS.Annex.IV.Par.8.coreTaxonomyReferenceModification",
                                  "ESEF.RTS.Annex.IV.Par.8.coreTaxonomyLabelModification"))
            del prohibitingLbElts[:]
            del prohibitedBaseConcepts[:]
        if not linkbasesFound and linkbaseRefType: # type of expected linkbase is known but it has no ext links
            # block top level warning message for 3.1.1 because reported here with specialized message
            linkEltName = None
            if linkbaseRefType == "cal":
                linkEltName = "calculationLink"
                val.hasExtensionCal = True
            elif linkbaseRefType == "def":
                linkEltName = "definitionLink"
                val.hasExtensionDef = True
            elif linkbaseRefType == "lab":
                linkEltName = "labelLink"
                val.hasExtensionLbl = True
            elif linkbaseRefType == "pre":
                linkEltName = "presentationLink"
                val.hasExtensionPre = True
            if linkEltName:
                linkbasesFound.add(linkEltName)
                val.modelXbrl.error("ESEF.3.1.1.extensionTaxonomyWrongFilesStructure",
                    _("Each linkbase type MUST be provided in a separate linkbase file: %(linkbaseType)s linkbase has no %(extendedLinkElement)s element."),
                    modelObject=modelDocument.xmlRootElement, linkbaseType=linkbaseRefType, extendedLinkElement=linkEltName)

        elif len(linkbasesFound) > 1:
            val.modelXbrl.error("ESEF.3.1.1.linkbasesNotSeparateFiles",
                _("Each linkbase type MUST be provided in a separate linkbase file, found: %(linkbasesFound)s."),
                modelObject=modelDocument.xmlRootElement, linkbasesFound=", ".join(sorted(linkbasesFound)))

        # check for any prohibiting dimensionArc's
        for prohibitingArcElt in modelDocument.xmlRootElement.iterdescendants(tag="{http://www.xbrl.org/2003/linkbase}definitionArc"):
            if (prohibitingArcElt.get("use") == "prohibited" and
                prohibitingArcElt.get("{http://www.w3.org/1999/xlink}arcrole")  == XbrlConst.dimensionDefault):
                val.modelXbrl.error("ESEF.3.4.3.extensionTaxonomyOverridesDefaultMembers",
                    _("The extension taxonomy MUST not prohibit default members assigned to dimensions by the ESEF taxonomy."),
                    modelObject=modelDocument.xmlRootElement, linkbasesFound=", ".join(sorted(linkbasesFound)))

    if isExtensionDoc and filenamePattern is not None:
        m = re.compile(filenameRegex).match(modelDocument.basename)
        if not m:
            val.modelXbrl.warning("ESEF.3.1.5.extensionTaxonomyDocumentNameDoesNotFollowNamingConvention",
                _("%(fileType)s file name SHOULD match the %(pattern)s pattern: %(documentName)s."),
                modelObject=modelDocument.xmlRootElement,
                fileType="Report" if modelDocument.type == ModelDocumentFile.Type.INLINEXBRL else "Extension taxonomy",
                pattern=filenamePattern, documentName=modelDocument.basename)
        elif len(m.group(1)) > 20:
            val.modelXbrl.warning("ESEF.3.1.5.baseComponentInNameOfTaxonomyFileExceedsTwentyCharacters",
                _("Extension taxonomy document file name {base} component SHOULD be no longer than 20 characters, length is %(length)s:  %(documentName)s."),
                modelObject=modelDocument.xmlRootElement, length=len(m.group(1)), documentName=modelDocument.basename)

    if isExtensionDoc and val.authority and val.authority.startswith("UKFRC"):
        if modelDocument.type == ModelDocumentFile.Type.INLINEXBRL:
            if modelDocument.documentEncoding.lower() != "utf-8":
                val.modelXbrl.error("UKFRC.1.1.instanceDocumentEncoding",
                    _("UKFRC instance documents should be UTF-8 encoded: %(encoding)s"),
                    modelObject=modelDocument, encoding=modelDocument.documentEncoding)



from __future__ import annotations

import itertools
import os
import re
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Literal


CONFORMANCE_SUITE_PATH_PREFIX = 'tests/resources/conformance_suites'


class AssetSource(Enum):
    LOCAL = 1
    S3_PUBLIC = 2
    S3_PRIVATE = 3

    def is_s3(self) -> bool:
        return self in (AssetSource.S3_PUBLIC, AssetSource.S3_PRIVATE)


class AssetType(Enum):
    CONFORMANCE_SUITE = 1
    CACHE_PACKAGE = 2
    TAXONOMY_PACKAGE = 3


@dataclass(frozen=True)
class ConformanceSuiteAssetConfig:
    local_filename: Path
    source: AssetSource
    type: AssetType
    extract_sequence: tuple[tuple[Path, Path], ...] = field(default_factory=tuple)
    entry_point: Path | None = field(compare=False, default=None)
    entry_point_root: Path | None = None
    public_download_url: str | None = None
    s3_key: str | None = None
    s3_version_id: str | None = None

    def __post_init__(self) -> None:
        if self.source.is_s3():
            assert self.s3_key is not None, 'Must provide S3 key for S3 assets.'
        else:
            assert self.s3_key is None, \
                'S3 key must not be provided for non-S3 assets.'
            assert self.s3_version_id is None, \
                'S3 version ID must not be provided for non-S3 assets.'
        assert bool(self.entry_point) == bool(self.entry_point_root),\
            'Entry point and entry point root must be both None or both set.'

    @cached_property
    def local_directory(self) -> Path:
        if self.type == AssetType.CONFORMANCE_SUITE:
            return Path('tests/resources/conformance_suites')
        if self.type == AssetType.TAXONOMY_PACKAGE:
            return Path('tests/resources/packages')
        return Path('tests/resources')

    @cached_property
    def full_entry_point(self) -> Path | None:
        if not self.full_entry_point_root or not self.entry_point:
            return None
        return self.full_entry_point_root / self.entry_point

    @cached_property
    def full_entry_point_root(self) -> Path | None:
        if not self.entry_point_root:
            return None
        return self.local_directory / self.entry_point_root

    @cached_property
    def full_local_path(self) -> Path:
        return self.local_directory / self.local_filename

    def get_conflicting_directories(self, reserved_directories: set[Path]) -> dict[Path, set[Path]]:
        return {
            k: v
            for k, v in {
                reserved_path: {
                    d for d in reserved_directories if d in reserved_path.parents
                }
                for reserved_path in self.full_reserved_paths
            }.items()
            if v
        }

    def get_conflicting_paths(self, reserved_paths: set[Path]) -> set[Path]:
        return reserved_paths.intersection(self.full_reserved_paths)

    @cached_property
    def full_reserved_directories(self) -> set[Path]:
        """
        :return: Set of directory paths this asset will write to.
        """
        return {
            self.local_directory / extract_to
            for __, extract_to in self.extract_sequence
        }

    @cached_property
    def full_reserved_paths(self) -> set[Path]:
        """
        :return: Set of file paths this asset will read from or write to.
        """
        return {
            self.full_local_path,
            self.full_entry_point_root or self.full_local_path,
        } | {
            self.local_directory / extract_from for extract_from, __ in self.extract_sequence
        }

    @staticmethod
    def cache_package(name: str, s3_version_id: str) -> ConformanceSuiteAssetConfig:
        return ConformanceSuiteAssetConfig(
            local_filename=Path(f'temp-{name}-cache.zip'),
            source=AssetSource.S3_PUBLIC,
            type=AssetType.CACHE_PACKAGE,
            s3_key=f'{name}.zip',
            s3_version_id=s3_version_id,
        )

    @staticmethod
    def conformance_suite(
            name: Path,
            entry_point: Path | None = None,
            public_download_url: str | None = None,
            source: AssetSource = AssetSource.S3_PRIVATE) -> ConformanceSuiteAssetConfig:
        return ConformanceSuiteAssetConfig(
            local_filename=name,
            source=source,
            type=AssetType.CONFORMANCE_SUITE,
            entry_point=entry_point,
            entry_point_root=name if entry_point else None,
            public_download_url=public_download_url,
            s3_key=name.as_posix() if source.is_s3() else None,
        )

    @staticmethod
    def local_conformance_suite(
            name: Path,
            entry_point: Path | None = None) -> ConformanceSuiteAssetConfig:
        return ConformanceSuiteAssetConfig(
            local_filename=name,
            source=AssetSource.LOCAL,
            type=AssetType.CONFORMANCE_SUITE,
            entry_point=entry_point,
            entry_point_root=name if entry_point else None,
        )

    @staticmethod
    def nested_conformance_suite(
            name: Path,
            extract_to: Path,
            entry_point_root: Path,
            entry_point: Path,
            public_download_url: str | None = None,
            source: AssetSource = AssetSource.S3_PRIVATE) -> ConformanceSuiteAssetConfig:
        return ConformanceSuiteAssetConfig(
            local_filename=name,
            source=source,
            type=AssetType.CONFORMANCE_SUITE,
            entry_point=entry_point,
            entry_point_root=entry_point_root,
            extract_sequence=(
                (name, extract_to),
            ),
            public_download_url=public_download_url,
            s3_key=name.as_posix() if source.is_s3() else None,
        )

    @staticmethod
    def public_taxonomy_package(name: Path, public_download_url: str | None = None) -> ConformanceSuiteAssetConfig:
        return ConformanceSuiteAssetConfig(
            local_filename=name,
            source=AssetSource.S3_PUBLIC,
            type=AssetType.TAXONOMY_PACKAGE,
            public_download_url=public_download_url,
            s3_key=name.as_posix(),
        )


@dataclass(frozen=True)
class ConformanceSuiteConfig:
    info_url: str
    name: str
    additional_plugins_by_prefix: list[tuple[str, frozenset[str]]] = field(default_factory=list)
    args: list[str] = field(default_factory=list)
    assets: list[ConformanceSuiteAssetConfig] = field(default_factory=list)
    cache_version_id: str | None = None
    capture_warnings: bool = True
    ci_enabled: bool = True
    expected_additional_testcase_errors: dict[str, frozenset[str]] = field(default_factory=dict)
    expected_failure_ids: frozenset[str] = frozenset()
    expected_missing_testcases: frozenset[str] = frozenset()
    expected_model_errors: frozenset[str] = frozenset()
    membership_url: str | None = None
    plugins: frozenset[str] = frozenset()
    shards: int = 1
    strict_testcase_index: bool = True
    url_replace: str | None = None
    network_or_cache_required: bool = True
    required_locale_by_ids: dict[str, re.Pattern[str]] = field(default_factory=dict)
    test_case_result_options: Literal['match-all', 'match-any'] = 'match-all'

    def __post_init__(self) -> None:
        redundant_plugins = [(prefix, overlap)
            for prefix, additional_plugins in self.additional_plugins_by_prefix
            for overlap in [self.plugins & additional_plugins]
            if overlap]
        assert not redundant_plugins, \
            f'Plugins specified both as default and additional: {redundant_plugins}'
        overlapping_prefixes = [(p1, p2)
            for (p1, _), (p2, _) in itertools.combinations(self.additional_plugins_by_prefix, 2)
            if p1.startswith(p2) or p2.startswith(p1)]
        assert not overlapping_prefixes, \
            f'Overlapping prefixes are not supported: {overlapping_prefixes}'
        assert not (self.shards == 1 and self.additional_plugins_by_prefix), \
            'Cannot specify additional_plugins_by_prefix with only one shard.'
        plugin_combinations = len({plugins for _, plugins in self.additional_plugins_by_prefix}) + 1
        assert plugin_combinations <= self.shards, \
            'Too few shards to accommodate the number of plugin combinations:' \
            f' combinations={plugin_combinations} shards={self.shards}'
        overlapping_expected_testcase_ids = self.expected_failure_ids.intersection(self.required_locale_by_ids)
        assert not overlapping_expected_testcase_ids, \
            f'Testcase IDs in both expected failures and required locales: {sorted(overlapping_expected_testcase_ids)}'
        overlapping_expected_failure_testcase_ids = self.expected_failure_ids.intersection(self.expected_additional_testcase_errors.keys())
        assert not overlapping_expected_failure_testcase_ids, \
            f'Testcase IDs in both expected failures and expected additional errors: {sorted(overlapping_expected_failure_testcase_ids)}'
        assert not self.network_or_cache_required or self.package_paths or self.cache_version_id, \
            'If network or cache is required, either packages must be used or a cache version ID must be provided.'

    @cached_property
    def entry_point_asset(self) -> ConformanceSuiteAssetConfig:
        entry_points = [asset for asset in self.assets if asset.entry_point]
        assert len(entry_points) == 1, \
            'Exactly one asset with entry point must be configured.'
        return entry_points[0]

    @cached_property
    def entry_point_path(self) -> Path:
        full_entry_point = self.entry_point_asset.full_entry_point
        assert full_entry_point is not None
        return full_entry_point

    @cached_property
    def entry_point_root(self) -> Path:
        entry_point_root = self.entry_point_asset.full_entry_point_root
        assert entry_point_root is not None
        return entry_point_root

    @cached_property
    def has_private_asset(self) -> bool:
        return any(asset.source == AssetSource.S3_PRIVATE for asset in self.assets)

    @cached_property
    def package_paths(self) -> set[Path]:
        return {
            asset.full_local_path
            for asset in self.assets
            if asset.type == AssetType.TAXONOMY_PACKAGE
        }



# This file is part of 'NTLM Authorization Proxy Server' http://sourceforge.net/projects/ntlmaps/
# Copyright 2001 Dmitry A. Rozmanov <dima@xenon.spb.ru>
#
# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.


C = 0x1000000000

def norm(n):
    return n & 0xFFFFFFFF


class U32:
    v = 0

    def __init__(self, value = 0):
        self.v = C + norm(abs(int(value)))

    def set(self, value = 0):
        self.v = C + norm(abs(int(value)))

    def __repr__(self):
        return hex(norm(self.v))

    def __long__(self): return int(norm(self.v))
    def __int__(self): return int(norm(self.v))
    def __chr__(self): return chr(norm(self.v))

    def __add__(self, b):
        r = U32()
        r.v = C + norm(self.v + b.v)
        return r

    def __sub__(self, b):
        r = U32()
        if self.v < b.v:
            r.v = C + norm(0x100000000 - (b.v - self.v))
        else: r.v = C + norm(self.v - b.v)
        return r

    def __mul__(self, b):
        r = U32()
        r.v = C + norm(self.v * b.v)
        return r

    def __div__(self, b):
        r = U32()
        r.v = C + (norm(self.v) / norm(b.v))
        return r

    def __mod__(self, b):
        r = U32()
        r.v = C + (norm(self.v) % norm(b.v))
        return r

    def __neg__(self): return U32(self.v)
    def __pos__(self): return U32(self.v)
    def __abs__(self): return U32(self.v)

    def __invert__(self):
        r = U32()
        r.v = C + norm(~self.v)
        return r

    def __lshift__(self, b):
        r = U32()
        r.v = C + norm(self.v << b)
        return r

    def __rshift__(self, b):
        r = U32()
        r.v = C + (norm(self.v) >> b)
        return r

    def __and__(self, b):
        r = U32()
        r.v = C + norm(self.v & b.v)
        return r

    def __or__(self, b):
        r = U32()
        r.v = C + norm(self.v | b.v)
        return r

    def __xor__(self, b):
        r = U32()
        r.v = C + norm(self.v ^ b.v)
        return r

    def __not__(self):
        return U32(not norm(self.v))

    def truth(self):
        return norm(self.v)

    def __cmp__(self, b):
        if norm(self.v) > norm(b.v): return 1
        elif norm(self.v) < norm(b.v): return -1
        else: return 0

    def __bool__(self):
        return norm(self.v)



from __future__ import annotations

import locale
import os.path
import re
import urllib.parse
from collections import Counter
from pathlib import PurePath
from typing import TYPE_CHECKING, cast

import pytest

from arelle import ModelDocument, PackageManager, PluginManager
from arelle.Cntlr import Cntlr
from arelle.CntlrCmdLine import parseAndRun
from arelle.FileSource import archiveFilenameParts

if TYPE_CHECKING:
    from _pytest.mark import ParameterSet


def get_document_id(doc: ModelDocument.ModelDocument) -> str:
    """
    Given a ModelDocument, attempt to find a basepath that can be used to generate a user-friendly document ID
    :param doc:
    :return: A parent path of the document's filepath.
    """
    parents = [p.as_posix() for p in PurePath(doc.filepath).parents]
    checked_paths = set()
    file_source = doc.modelXbrl.fileSource
    # Try basefile
    basefile: str = cast(str, getattr(file_source, 'basefile', None))
    if basefile is not None:
        basefile = PurePath(basefile).as_posix()
        if basefile in parents:
            return get_document_id_from_basepath(doc, basefile)
        else:
            checked_paths.add(basefile)
    # Try referenced documents
    ref_file_source = next(iter(file_source.referencedFileSources.values()), None)
    ref_basefile: str = ref_file_source.basefile if ref_file_source is not None else None
    if ref_basefile is not None:
        ref_basefile = PurePath(ref_basefile).as_posix()
        if ref_basefile in parents:
            return get_document_id_from_basepath(doc, ref_basefile)
        else:
            checked_paths.add(ref_basefile)
    # Try archive subpath
    archive_path_parts = archiveFilenameParts(doc.filepath)
    if archive_path_parts is not None:
        archive_path_part = PurePath(archive_path_parts[1]).as_posix()
        return archive_path_part
    # Use file source URL as fallback if basepath not found
    file_source_url = PurePath(os.path.dirname(file_source.url)).as_posix()
    if file_source_url in parents:
        return get_document_id_from_basepath(doc, file_source_url)
    else:
        checked_paths.add(file_source_url)
    raise ValueError(f'Could not determine basepath. '
                     f'None of the checked paths ({checked_paths}) were parents of \"{doc.filepath}\".')


def get_document_id_from_basepath(doc: ModelDocument.ModelDocument, basepath: str) -> str:
    return PurePath(doc.filepath).relative_to(basepath).as_posix()


def get_s3_uri(path: str, version_id: str | None = None) -> str:
    path = urllib.parse.quote(path)
    uri = f'https://arelle-public.s3.amazonaws.com/{path}'
    if version_id is not None:
        uri += f'?versionId={version_id}'
    return uri


def get_test_data(
        args: list[str],
        expected_failure_ids: frozenset[str] = frozenset(),
        expected_model_errors: frozenset[str] = frozenset(),
        required_locale_by_ids: dict[str, re.Pattern[str]] | None = None,
        strict_testcase_index: bool = True,
) -> list[ParameterSet]:
    """
    Produces a list of Pytest Params that can be fed into a parameterized pytest function

    :param args: The args to be parsed by arelle in order to correctly produce the desired result set
    :param expected_failure_ids: The set of string test IDs that are expected to fail
    :param expected_model_errors: The set of error codes expected to be in the ModelXbrl errors
    :param required_locale_by_ids: The dict of IDs for tests which require a system locale matching a regex pattern.
    :param strict_testcase_index: Don't allow IOerrors when loading the testcase index
    :return: A list of PyTest Params that can be used to run a parameterized pytest function
    """
    if required_locale_by_ids is None:
        required_locale_by_ids = {}
    cntlr = parseAndRun(args)  # type: ignore[no-untyped-call]
    try:
        system_locale = locale.setlocale(locale.LC_CTYPE)
        results: list[ParameterSet] = []
        test_cases_with_no_variations = set()
        test_cases_with_unrecognized_type = {}
        model_document = cntlr.modelManager.modelXbrl.modelDocument
        test_cases: list[ModelDocument.ModelDocument] = []
        if strict_testcase_index and model_document.type == ModelDocument.Type.TESTCASESINDEX:
            model_errors = sorted(cntlr.modelManager.modelXbrl.errors)
            assert 'IOerror' not in model_errors, f'One or more testcases referenced by testcases index "{model_document.filepath}" were not found.'
        collect_test_data(
            cntlr=cntlr,
            expected_failure_ids=expected_failure_ids,
            expected_model_errors=expected_model_errors,
            required_locale_by_ids=required_locale_by_ids,
            system_locale=system_locale,
            results=results,
            model_document=model_document,
            test_cases=test_cases,
        )
        for test_case in sorted(test_cases, key=lambda doc: doc.uri):
            test_case_file_id = get_document_id(test_case)
            if test_case.type not in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
                test_cases_with_unrecognized_type[test_case_file_id] = test_case.type
            if not getattr(test_case, "testcaseVariations", None):
                test_cases_with_no_variations.add(test_case_file_id)
            else:
                for mv in test_case.testcaseVariations:
                    test_id = f'{test_case_file_id}:{mv.id}'
                    marks = []
                    if isExpectedFailure(test_id, expected_failure_ids, required_locale_by_ids, system_locale):
                        marks.append(pytest.mark.xfail())
                    elif mv.status == 'skip':
                        continue  # don't report variations skipped due to shards
                    # Arelle adds message code frequencies to the end, but conformance suites usually don't.
                    # Skip assertion results dictionaries.
                    actual = [re.sub(r' \(\d+\)$', '', code) for code in mv.actual if not isinstance(code, dict)]
                    param = pytest.param(
                        {
                            'status': mv.status,
                            'expected': mv.expected,
                            'actual': actual,
                            'duration': mv.duration,
                        },
                        id=test_id,
                        marks=marks,
                    )
                    results.append(param)
        if test_cases_with_unrecognized_type:
            raise Exception(f"Some test cases have an unrecognized document type: {sorted(test_cases_with_unrecognized_type.items())}.")
        test_id_frequencies = Counter(cast(str, p.id) for p in results)
        nonunique_test_ids = {test_id: count for test_id, count in test_id_frequencies.items() if count > 1}
        if nonunique_test_ids:
            raise Exception(f'Some test IDs are not unique.  Frequencies of nonunique test IDs: {nonunique_test_ids}.')
        nonexistent_expected_failure_ids = expected_failure_ids - test_id_frequencies.keys()
        if nonexistent_expected_failure_ids:
            raise Exception(f"Some expected failure IDs don't match any test cases: {sorted(nonexistent_expected_failure_ids)}.")
        nonexistent_required_locale_testcase_ids = required_locale_by_ids.keys() - test_id_frequencies.keys()
        if nonexistent_required_locale_testcase_ids:
            raise Exception(f"Some required locale IDs don't match any test cases: {sorted(nonexistent_required_locale_testcase_ids)}.")
        return results
    finally:
        cntlr.modelManager.close()
        PackageManager.close()  # type: ignore[no-untyped-call]
        PluginManager.close()  # type: ignore[no-untyped-call]


def collect_test_data(
        cntlr: Cntlr,
        expected_failure_ids: frozenset[str],
        expected_model_errors: frozenset[str],
        required_locale_by_ids: dict[str, re.Pattern[str]],
        system_locale: str,
        results: list[ParameterSet],
        model_document: ModelDocument.ModelDocument,
        test_cases: list[ModelDocument.ModelDocument],
) -> None:
    if model_document.type == ModelDocument.Type.TESTCASESINDEX:
        for child_document in model_document.referencesDocument.keys():
            collect_test_data(
                cntlr=cntlr,
                expected_failure_ids=expected_failure_ids,
                expected_model_errors=expected_model_errors,
                required_locale_by_ids=required_locale_by_ids,
                system_locale=system_locale,
                results=results,
                model_document=child_document,
                test_cases=test_cases,
            )
    elif model_document.type == ModelDocument.Type.REGISTRY:
        test_cases.extend(model_document.referencesDocument.keys())
    elif model_document.type in (ModelDocument.Type.TESTCASE, ModelDocument.Type.REGISTRYTESTCASE):
        test_cases.append(model_document)
    elif model_document.type == ModelDocument.Type.INSTANCE:
        test_id = get_document_id(model_document)
        expected_failure = isExpectedFailure(test_id, expected_failure_ids, required_locale_by_ids, system_locale)
        model_errors = sorted(cntlr.modelManager.modelXbrl.errors)
        expected_model_errors_list = sorted(expected_model_errors)
        param = pytest.param(
            {
                'status': 'pass' if model_errors == expected_model_errors_list else 'fail',
                'expected': expected_model_errors_list,
                'actual': model_errors,
            },
            id=test_id,
            marks=[pytest.mark.xfail()] if expected_failure else [],
        )
        results.append(param)
    else:
        raise Exception('Unhandled model document type: {}'.format(model_document.type))


def isExpectedFailure(
        test_id: str,
        expected_failure_ids: frozenset[str],
        required_locale_by_ids: dict[str, re.Pattern[str]],
        system_locale: str,
) -> bool:
    if test_id in expected_failure_ids:
        return True
    if test_id in required_locale_by_ids:
        return not required_locale_by_ids[test_id].search(system_locale)
    return False



# -*- coding: utf-8 -*-
"""
:mod:`EdgarFtIxGen.FtJsonToOimJson`
~~~~~~~~~~~~~~~~~~~
FtJsonToOimJson was created by staff of the U.S. Securities and Exchange Commission.
Data and content created by government employees within the scope of their employment
are not subject to domestic copyright protection. 17 U.S.C. 105.

This plugin can be run as a main program (with file name argument for FT JSON input) or as a
report setup plugin to LoadFromOIM (in which case it converts the FT JSON into OIM JSON
during the LoadFromOIM process).

"""

import datetime, json, re, os

unitByName = (
        ("iso4217:USD", re.compile("(?!.*[Ee]xpltn).*(Amt|Val|Pric|Valtn|PdFee|PricFsclYr|PricPrrFsclYr|Cdts)$|(?!.*Per).*([vV]al)[A-Z]|(csh|val).*$")),
        ("xbrli:shares", re.compile(".*(NbOf).*$|(amtScties).*$")),
        (None, re.compile(".*Rate.*$")) # pure is omitted
    )

def unitForName(name):
    for unit, namePattern in unitByName:
        if namePattern.match(name):
            return unit
    return None

def prefixForName(name):
    if name in ("entityCentralIndexKey", "entityRegistrantName"):
        return "dei"
    return "ffd"

def getLatestTaxonomyFamily(modelXbrl, name):
    latest_version = None
    if not modelXbrl:
        # keeping this hard coded version just for debugging purposes.
        # This plugin is called via loadFromOIM plugin so it should have
        # a modelXbrl when called.
        class taxonomyFamily:
            def __init__(self, name):
                self.family = name.lower()
                self.namespace = f"http://xbrl.sec.gov/{self.family}/2024"
                self.href = f"https://xbrl.sec.gov/{self.family}/2024/{self.family}-2024.xsd"
        latest_version = taxonomyFamily(name)
    else:
        for family in modelXbrl.modelManager.disclosureSystem.familyHrefs.get(name, []):
            prefix = family.namespace.split("/")[-2]
            if not latest_version and prefix == name.lower():
                latest_version = family
            else:
                if prefix == name.lower() and family.version > latest_version.version:
                    latest_version = family
    return latest_version

def ftJsonToOimJson(modelXbrl, oimObject, *args, **kwargs):
    # is this an FT JSON file?
    if not( isinstance(oimObject, dict) and "documentInfo" not in oimObject and all(
            t in oimObject for t in ("feesSummaryTable", "submissionTable"))):
        return None
    ftJson = oimObject
    CIK = oimObject.get("submissionTable", {}).get("entityCentralIndexKey", "0000000000")
    today = datetime.datetime.today()
    start_date = (today - datetime.timedelta(days=1)).strftime("%Y-%m-%d")
    end_date = today.strftime("%Y-%m-%d")
    period = f"{start_date}T00:00:00/{end_date}T00:00:00"
    facts = {}
    dei = getLatestTaxonomyFamily(modelXbrl, "DEI")
    ffd = getLatestTaxonomyFamily(modelXbrl, "FFD")
    ftOim = {
        "documentInfo": {
            "documentType": "https://xbrl.org/2021/xbrl-json",
            "features": {},
            "namespaces": {
                "dei": dei.namespace,
                "ffd": ffd.namespace,
                "iso4217": "http://www.xbrl.org/2003/iso4217",
                "scheme": "http://www.sec.gov/CIK",
                "xbrli": "http://www.xbrl.org/2003/instance",
                "xbrl": "https://xbrl.org/2021"
            },
            "taxonomy": [
                ffd.href
            ]
        },
        "facts": facts
    }

    ignoreZeroFlagNamePattern = re.compile(r"ffd:(Rule(011|457[aoursf]|415a6|457[bp]Offset|011a2Offset)Flg|FeesOthrRuleFlg)")

    # relative import of sibling module Util is iffy in cx_frozen builds, read ft-validations directly
    with open(os.path.join(os.path.dirname(__file__),"resources","ft-validations.json"), "r", encoding="utf-8") as fh:
        ftValidations = json.load(fh)
        formFields =  ftValidations['form-fields']
        formCodeMapping = {v: k for k, v in ftValidations['form-mapping'].items()}

    def newFact(name, value, axis=None, tableLine=None):
        unit = unitForName(name)
        if ":" not in name:
            name = f"{prefixForName(name)}:{name[0].upper()}{name[1:]}"
        if isinstance(value, (bool,int,float)):
            if unit == "xbrli:shares" and isinstance(value, float):
                value = int(value)
            value = str(value).lower() # OIM requires all fact values be string in OIM-JSON input
        if value in ("0", "false") and ignoreZeroFlagNamePattern.match(name):
            return # ignore flag
        if name in formFields and value in formCodeMapping:
            value = formCodeMapping[value]
        facts[f"f{len(facts)+1}"] = fact= {
            "value": value,
            "dimensions": {
                "concept": name,
                "entity": f"scheme:{CIK}",
                "period": period
            }
        }
        if unit:
            fact["dimensions"]["unit"] = unit
        if axis:
            fact["dimensions"][axis] = f"{tableLine}"

    for table, axis in (("feesSummaryTable", None),
                        ("offeringTable", "ffd:OfferingAxis"),
                        ("offsetTable", "ffd:OffsetAxis"),
                        ("cmbndPrspctsTable", "ffd:CmbndPrspctsItemAxis"),
                        ("submissionTable", None),
                        ("scties424iTable", "ffd:Scties424iAxis")
                        ):
        tableLine = 1
        lines = ftJson.get(table, [])
        if axis is None and isinstance(lines, dict): # top level object only
            lines = [lines]
        for line in lines:
            for name, value in line.items():
                if not table.startswith(name) and not name.endswith("InnerText") and not name in {"!@%duplicateValues%@!", "tpOfPmt", "nbDaysLate", "cmbndPrspctsItem"} and \
                not (axis=="ffd:Scties424iAxis" and name=="rule457uFlg"): # rule457uFlg is derived for EDGAR on 424i subtypes but not part of taxonomy
                    newFact(name, value, axis, tableLine)
            tableLine += 1

    oimObject.clear()
    oimObject.update(ftOim)
    #with open("/Users/hermf/Documents/mvsl/projects/SEC/FT/LN-samples/457bOffset1Source-test.json", "w", encoding="utf-8") as fh:
    #    fh.write(json.dumps(oimObject, indent=3))

    if modelXbrl is not None:
        modelXbrl.loadedFromFtJson = True
    return ftOim

def isFtJsonDocument(modelXbrl, *args, **kwargs):
    return getattr(modelXbrl, "loadedFromFtJson", False)


# debugging
if __name__ == "__main__":
    with open(r"C:\Users\Lopezfr\Documents\AB\EXFILINGFEES.json", "rt", encoding="utf-8") as fh:
        ftJson = json.load(fh)
    oim = ftJsonToOimJson(None, ftJson)
    with open(r"C:\Users\Lopezfr\Documents\AB\EXFILINGFEESoim.json", "w", encoding="utf-8") as fh:
        fh.write(json.dumps(oim, indent=3))

__pluginInfo__ = {
    'name': 'FT JSON to OIM JSON',
    'version': '1.0',
    'description': "This plug-in converts an FT JSON document into an xBRL-JSON OIM report.",
    'license': 'Apache-2',
    'author': 'SEC Employees',
    'copyright': '(c) Portions by SEC Employees not subject to domestic copyright.',
    'import': ('loadFromOIM', ), # import dependent modules
    # classes of mount points (required)
    'LoadFromOim.DocumentSetup': ftJsonToOimJson,
    'FtJson.IsFtJsonDocument': isFtJsonDocument
}



# This library is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.  If not, see <http://www.gnu.org/licenses/> or <http://www.gnu.org/licenses/lgpl.txt>.

import urllib.request, urllib.error
import http.client, socket
from urllib.response import addinfourl
try:
    from . import ntlm
except ValueError:
    import ntlm
import regex as re

class AbstractNtlmAuthHandler:

    def __init__(self, password_mgr=None):
        if password_mgr is None:
            password_mgr = HTTPPasswordMgr()
        self.passwd = password_mgr
        self.add_password = self.passwd.add_password

    def http_error_authentication_required(self, auth_header_field, req, fp, headers):
        auth_header_value_list = headers.get_all(auth_header_field)
        if auth_header_value_list:
            if any([hv.lower() == 'ntlm' for hv in auth_header_value_list]):
                fp.close()
                return self.retry_using_http_NTLM_auth(req, auth_header_field, None, headers)

    def retry_using_http_NTLM_auth(self, req, auth_header_field, realm, headers):
        user, pw = self.passwd.find_user_password(realm, req.get_full_url())
        if pw is not None:
            user_parts = user.split('\\', 1)
            if len(user_parts) == 1:
                UserName = user_parts[0]
                DomainName = ''
                type1_flags = ntlm.NTLM_TYPE1_FLAGS & ~ntlm.NTLM_NegotiateOemDomainSupplied
            else:
                DomainName = user_parts[0].upper()
                UserName = user_parts[1]
                type1_flags = ntlm.NTLM_TYPE1_FLAGS
            # ntlm secures a socket, so we must use the same socket for the complete handshake
            headers = dict(req.headers)
            headers.update(req.unredirected_hdrs)
            auth = 'NTLM %s' % ntlm.create_NTLM_NEGOTIATE_MESSAGE(
                user, type1_flags
            ).decode('ascii')
            if req.headers.get(self.auth_header, None) == auth:
                return None
            headers[self.auth_header] = auth

            host = req.host
            if not host:
                raise urllib.request.URLError('no host given')
            h = None
            if req.get_full_url().startswith('https://'):
                h = http.client.HTTPSConnection(host) # will parse host:port
            else:
                h = http.client.HTTPConnection(host) # will parse host:port
            # we must keep the connection because NTLM authenticates the connection, not single requests
            headers["Connection"] = "Keep-Alive"
            headers = dict((name.title(), val) for name, val in list(headers.items()))
            h.request(req.get_method(), req.selector, req.data, headers)
            r = h.getresponse()
            r.begin()
            r._safe_read(int(r.getheader('content-length')))
            try:
                if r.getheader('set-cookie'):
                    # this is important for some web applications that store authentication-related info in cookies (it took a long time to figure out)
                    headers['Cookie'] = r.getheader('set-cookie')
            except TypeError:
                pass
            r.fp = None # remove the reference to the socket, so that it can not be closed by the response object (we want to keep the socket open)
            auth_header_value = r.getheader(auth_header_field, None)

            # some Exchange servers send two WWW-Authenticate headers, one with the NTLM challenge
            # and another with the 'Negotiate' keyword - make sure we operate on the right one
            m = re.match(r'(NTLM [A-Za-z0-9+\-/=]+)', auth_header_value)
            if m:
                auth_header_value, = m.groups()

            (ServerChallenge, NegotiateFlags) = ntlm.parse_NTLM_CHALLENGE_MESSAGE(auth_header_value[5:])
            auth = 'NTLM %s' % ntlm.create_NTLM_AUTHENTICATE_MESSAGE(
                ServerChallenge, UserName, DomainName, pw, NegotiateFlags
            ).decode('ascii')
            headers[self.auth_header] = auth
            headers["Connection"] = "Close"
            headers = dict((name.title(), val) for name, val in list(headers.items()))
            try:
                h.request(req.get_method(), req.selector, req.data, headers)
                # none of the configured handlers are triggered, for example redirect-responses are not handled!
                response = h.getresponse()
                def notimplemented():
                    raise NotImplementedError
                response.readline = notimplemented
                return addinfourl(response, response.msg, req.get_full_url(), response.code)
            except socket.error as err:
                raise urllib.request.URLError(err)
        else:
            return None


class HTTPNtlmAuthHandler(AbstractNtlmAuthHandler, urllib.request.BaseHandler):

    auth_header = 'Authorization'

    def http_error_401(self, req, fp, code, msg, headers):
        return self.http_error_authentication_required('www-authenticate', req, fp, headers)


class ProxyNtlmAuthHandler(AbstractNtlmAuthHandler, urllib.request.BaseHandler):
    """
        CAUTION: this class has NOT been tested at all!!!
        use at your own risk
    """
    auth_header = 'Proxy-authorization'

    def http_error_407(self, req, fp, code, msg, headers):
        return self.http_error_authentication_required('proxy-authenticate', req, fp, headers)


if __name__ == "__main__":
    url = "http://ntlmprotectedserver/securedfile.html"
    user = "DOMAIN\\User"
    password = "Password"
    passman = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    passman.add_password(None, url, user , password)
    auth_basic = urllib.request.HTTPBasicAuthHandler(passman)
    auth_digest = urllib.request.HTTPDigestAuthHandler(passman)
    auth_NTLM = HTTPNtlmAuthHandler(passman)

    # disable proxies (just for testing)
    proxy_handler = urllib.request.ProxyHandler({})

    opener = urllib.request.build_opener(proxy_handler, auth_NTLM) #, auth_digest, auth_basic)

    urllib.request.install_opener(opener)

    response = urllib.request.urlopen(url)
    print((response.read()))



from __future__ import annotations

import boto3
import os
import urllib.request
import zipfile
from pathlib import Path

from tests.integration_tests.download_cache import apply_cache
from tests.integration_tests.integration_test_util import get_s3_uri
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteAssetConfig, AssetType, AssetSource


def _download_asset(asset: ConformanceSuiteAssetConfig, download_private: bool) -> bool:
    """
    Given an asset, determines the source to attempt download from,
    then performs the download.
    If the asset is configured with a private S3 source, will only attempt
    download if `download_private` is `True`, otherwise it will fall back
    to a public download URL, if configured.
    If not configured, no download will be performed and `False` returned.
    :return: True if download was successful.
    """
    if asset.source == AssetSource.LOCAL:
        print(f'Using local asset: {asset.full_local_path}')
        return True
    if asset.source == AssetSource.S3_PUBLIC:
        print(f'Using public S3 asset: {asset.full_local_path}')
        _download_public_s3_asset(asset)
        return True
    if asset.source == AssetSource.S3_PRIVATE and download_private:
        print(f'Using private S3 asset: {asset.full_local_path}')
        _download_private_s3_asset(asset)
        return True
    # Asset wasn't otherwise downloaded, fallback to public download URL, if configured.
    if asset.public_download_url:
        print(f'Using public asset: {asset.full_local_path}')
        _download_public_uri(asset.public_download_url, asset.full_local_path)
        return True
    return False


def _download_private_s3_asset(asset: ConformanceSuiteAssetConfig) -> None:
    """
    Downloads given asset from private S3 bucket.
    Will raise an error if the required AWS
    environment variables are not set.
    """
    key = asset.s3_key
    assert key is not None
    if asset.type == AssetType.CONFORMANCE_SUITE:
        key = f'conformance_suites/{key}'
    asset.full_local_path.parent.mkdir(parents=True, exist_ok=True)
    assert 'AWS_ACCESS_KEY_ID' in os.environ.keys(), 'Must have AWS_ACCESS_KEY_ID environment variable set.'
    assert 'AWS_SECRET_ACCESS_KEY' in os.environ.keys(), 'Must have AWS_SECRET_ACCESS_KEY environment variable set.'
    s3 = boto3.client('s3')
    if asset.s3_version_id:
        s3.download_file('arelle', Key=key, Filename=str(asset.full_local_path), ExtraArgs={'VersionId': asset.s3_version_id})
    else:
        s3.download_file('arelle', Key=key, Filename=str(asset.full_local_path))


def _download_public_s3_asset(asset: ConformanceSuiteAssetConfig) -> None:
    """
    Downloads given asset from public S3 bucket.
    """
    key = asset.s3_key
    assert key is not None
    if asset.type == AssetType.CONFORMANCE_SUITE:
        key = f'ci/conformance_suites/{key}'
    elif asset.type == AssetType.CACHE_PACKAGE:
        key = f'ci/caches/conformance_suites/{key}'
    elif asset.type == AssetType.TAXONOMY_PACKAGE:
        key = f'ci/taxonomy_packages/{key}'
    uri = get_s3_uri(key, version_id=asset.s3_version_id)
    _download_public_uri(uri, asset.full_local_path)


def _download_public_uri(uri: str, download_path: Path) -> None:
    """
    Downloads asset from given URI to given download path.
    """
    try:
        download_path.parent.mkdir(parents=True, exist_ok=True)
        urllib.request.urlretrieve(uri, download_path)
    except Exception as exc:
        raise Exception(f'Failed to download public URI "{uri}" to "{download_path}".') from exc


def _extract_asset(asset: ConformanceSuiteAssetConfig) -> None:
    """
    Performs the asset's configured extract sequence.
    An extract sequence is a list of from/to path pairs.
    For each element in the sequence, the zip at the `from` path
    is extracted to the `to` directory.
    """
    for extract_from, extract_to in asset.extract_sequence:
        extract_from = Path('tests/resources/conformance_suites') / extract_from
        extract_to = Path('tests/resources/conformance_suites') / extract_to
        with zipfile.ZipFile(extract_from, 'r') as zip_ref:
            os.makedirs(extract_to, exist_ok=True)
            zip_ref.extractall(extract_to)


def download_assets(
    assets: set[ConformanceSuiteAssetConfig],
    overwrite: bool,
    download_and_apply_cache: bool,
    download_private: bool,
) -> None:
    """
    Performs download for each of the given assets. Before each download, the asset's set of
    reserved paths is compared against the reserved paths of previous assets. If any conflicts
    are found (indicating two different assets may be overwriting each other), an `AssertionError`
    is raised.
    :param assets:
    :param overwrite: If assets that appear to be already downloaded should be skipped.
    :param download_and_apply_cache: If cache-type assets should be downloaded and applied or skipped.
    :param download_private: If download of private-source assets should be attempted.
    """
    reserved_paths: set[Path] = set()
    reserved_directories: set[Path] = set()
    for asset in assets:
        conflicting_paths = asset.get_conflicting_paths(reserved_paths)
        assert not conflicting_paths, \
            f'Assets have conflicting paths: {conflicting_paths}'
        conflicting_directories = asset.get_conflicting_directories(reserved_directories)
        assert not conflicting_directories, \
            f'Asset has reserved paths that conflict ' \
            f'with reserved directories: {conflicting_directories}'
        reserved_paths.update(asset.full_reserved_paths)
        reserved_directories.update(asset.full_reserved_directories)
        if os.path.exists(asset.full_local_path) and not overwrite:
            print(f'Using existing local asset: {asset.full_local_path}')
            continue  # File already exists at location, overwrite not specified
        if asset.type == AssetType.CACHE_PACKAGE and not download_and_apply_cache:
            continue  # Cache download not requested
        downloaded = _download_asset(asset, download_private)
        if downloaded:
            _extract_asset(asset)
        verify_asset(asset)
        if asset.type == AssetType.CACHE_PACKAGE:
            apply_cache(str(asset.full_local_path))
            print(f'Applied cache: {asset.full_local_path}')


def verify_asset(asset: ConformanceSuiteAssetConfig) -> None:
    """
    Verifies that an asset exists at it's expected location.
    For entry point assets, verifies that the entry point exists where it is expected.
    Raises an `AssertionError` on failure, otherwise logs a message.
    """
    assert asset.full_local_path.exists(), \
        f'Missing asset: {asset}'
    if asset.full_entry_point_root:
        if zipfile.is_zipfile(asset.full_entry_point_root):
            with zipfile.ZipFile(asset.full_entry_point_root, 'r') as zip_ref:
                assert asset.entry_point and asset.entry_point.as_posix() in zip_ref.namelist(), \
                    f'Asset entry point {asset.entry_point} does not exist in archive: {asset.full_entry_point_root}'
        else:
            assert asset.full_entry_point and asset.full_entry_point.exists(), \
                f'Asset entry point does not exist: {asset.full_entry_point}'
    print(f'Verified asset: {asset.full_local_path}')


def verify_assets(assets: set[ConformanceSuiteAssetConfig]) -> None:
    """
    Verifies each of the given assets exists locally.
    Raises an `AssertionError` on failure.
    Skips cache assets as they are deleted once applied.
    """
    for asset in assets:
        if asset.type == AssetType.CACHE_PACKAGE:
            continue  # Cache packages are deleted after being applied
        verify_asset(asset)



'''
See COPYRIGHT.md for copyright information.
'''
import os, io, json
from collections import OrderedDict
from arelle import ViewFile
from arelle.ModelObject import ModelObject
from arelle.Aspect import Aspect, aspectModels, aspectRuleAspects, aspectModelAspect, aspectStr
from arelle.ModelInstanceObject import ModelDimensionValue
from arelle.ModelRenderingObject import (StrctMdlNode, StrctMdlTable, StrctMdlBreakdown, StrctMdlStructuralNode,
                                         OPEN_ASPECT_ENTRY_SURROGATE, ROLLUP_SPECIFIES_MEMBER, ROLLUP_IMPLIES_DEFAULT_MEMBER,
                                         ROLLUP_FOR_CONCEPT_RELATIONSHIP_NODE, ROLLUP_FOR_DIMENSION_RELATIONSHIP_NODE,
                                         ROLLUP_FOR_CLOSED_DEFINITION_NODE, ROLLUP_FOR_OPEN_DEFINITION_NODE,
                                         ROLLUP_FOR_DEFINITION_NODE)
from arelle.rendering.RenderingResolution import resolveTableStructure
from arelle.rendering.RenderingLayout import layoutTable
from arelle.ModelValue import QName
from arelle import XbrlConst

def viewRenderedStructuralModel(modelXbrl, outfile, lang=None, viewTblELR=None, sourceView=None, cssExtras=""):
    modelXbrl.modelManager.showStatus(_("saving rendered structure"))
    view = ViewRenderedStructuralModel(modelXbrl, outfile, lang, cssExtras)
    view.view(outfile, viewTblELR)
    view.close(noWrite=True) # written out below
    modelXbrl.modelManager.showStatus(_("rendering table saved to {0}").format(outfile), clearAfter=5000)

class ViewRenderedStructuralModel(ViewFile.View):
    def __init__(self, modelXbrl, outfile, lang, cssExtras):
        # find table model namespace based on table namespace
        self.tableModelNamespace = XbrlConst.tableModel
        for xsdNs in modelXbrl.namespaceDocs.keys():
            if xsdNs in (XbrlConst.tableMMDD, XbrlConst.table):
                self.tableModelNamespace = xsdNs + "/model"
                break
        super(ViewRenderedStructuralModel, self).__init__(modelXbrl, outfile,
                                               'dummyObject',
                                               lang,
                                               style="rendering",
                                               cssExtras=cssExtras)
        class nonTkBooleanVar():
            def __init__(self, value=True):
                self.value = value
            def set(self, value):
                self.value = value
            def get(self):
                return self.value
        # context menu boolean vars (non-tkinter boolean
        self.ignoreDimValidity = nonTkBooleanVar(value=True)


    def tableModelQName(self, localName):
        return '{' + self.tableModelNamespace + '}' + localName

    def viewReloadDueToMenuAction(self, *args):
        self.view()

    def view(self, outfile, viewTblELR=None):
        if viewTblELR is not None:
            tblELRs = (viewTblELR,)
        else:
            tblELRs = self.modelXbrl.relationshipSet("Table-rendering").linkRoleUris

        for tblELR in tblELRs:
            self.zOrdinateChoices = {}

            strctMdlTable = resolveTableStructure(self, tblELR)

            # uncomment below for debugging Definition and Structural Models
            def jsonStrctMdlEncoder(obj, indent="\n"):
                if isinstance(obj, StrctMdlNode):
                    o = OrderedDict()
                    o["object"] = obj.__repr__()
                    if obj.xlinkLabel is not None:
                        o["defnMdlNode"] = f"{obj.defnMdlNode.modelXbrl.modelDocument.basename} line {obj.defnMdlNode.sourceline} {obj.xlinkLabel}"
                    if isinstance(obj, StrctMdlTable):
                        o["entryFile"] = obj.defnMdlNode.modelXbrl.modelDocument.basename,
                    if obj.axis:
                        o["axis"] = obj.axis
                    if obj.isAbstract:
                        o["abstract"] = True
                    if isinstance(obj, StrctMdlStructuralNode):
                        if obj.hasChildRollup:
                            o["hasChildRollup"] = True
                        if obj.rollup:
                            o["rollup"] = {ROLLUP_SPECIFIES_MEMBER:"rollup specifies member",
                                           ROLLUP_IMPLIES_DEFAULT_MEMBER:"rollup implies default member",
                                           ROLLUP_FOR_CONCEPT_RELATIONSHIP_NODE:"rollup for concept relationship nesting",
                                           ROLLUP_FOR_DIMENSION_RELATIONSHIP_NODE:"rollup for concept relationship nesting",
                                           ROLLUP_FOR_CLOSED_DEFINITION_NODE:"rollup for closed definition node",
                                           ROLLUP_FOR_OPEN_DEFINITION_NODE:"rollup for open definition node",
                                           ROLLUP_FOR_DEFINITION_NODE:"rollup for definition node"}[obj.rollup]
                        o["structuralDepth"] = obj.structuralDepth
                        _aspectsCovered = obj.aspectsCovered()
                        if _aspectsCovered:
                            o["aspectsCovered"] = OrderedDict((aspectStr(a),
                                                               str(v.stringValue if isinstance(v,ModelObject) else v
                                                                   ).replace(OPEN_ASPECT_ENTRY_SURROGATE, "OPEN_ASPECT_ENTRY_"))
                                                              for a in _aspectsCovered
                                                              if a != Aspect.DIMENSIONS
                                                              for v in (obj.aspectValue(a),))
                    if obj.tagSelector:
                        o["tagSelector"] = obj.tagSelector
                    if obj.strctMdlChildNodes:
                        o["strctMdlChildNodes"] = obj.strctMdlChildNodes
                    return o
                raise TypeError("Type {} is not supported for json output".format(type(obj).__name__))
            with io.open(outfile, 'wt') as fh:
                json.dump(strctMdlTable, fh, ensure_ascii=False, indent=2, default=jsonStrctMdlEncoder)



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import simpledialog, Toplevel, font, messagebox, VERTICAL, HORIZONTAL, N, S, E, W
from tkinter.constants import DISABLED, ACTIVE
try:
    from tkinter.ttk import Treeview, Scrollbar, Frame, Label, Button
except ImportError:
    from ttk import Treeview, Scrollbar, Frame, Label, Button
from arelle import PackageManager, DialogURL, DialogOpenArchive
from arelle.CntlrWinTooltip import ToolTip
import os, time, json, sys, traceback
import regex as re

STANDARD_PACKAGES_URL = "https://taxonomies.xbrl.org/api/v0/taxonomy"

def dialogPackageManager(mainWin):
    # check for updates in background
    import threading
    thread = threading.Thread(target=lambda cntlr=mainWin: backgroundCheckForUpdates(cntlr))
    thread.daemon = True
    thread.start()

def backgroundCheckForUpdates(cntlr):
    cntlr.showStatus(_("Checking for updates to packages")) # clear web loading status
    packageNamesWithNewerFileDates = PackageManager.packageNamesWithNewerFileDates()
    if packageNamesWithNewerFileDates:
        cntlr.showStatus(_("Updates are available for these packages: {0}")
                              .format(', '.join(packageNamesWithNewerFileDates)), clearAfter=5000)
    else:
        cntlr.showStatus(_("No updates found for packages."), clearAfter=5000)
    time.sleep(0.1) # Mac locks up without this, may be needed for empty ui queue?
    cntlr.uiThreadQueue.put((DialogPackageManager, [cntlr, packageNamesWithNewerFileDates]))

class DialogPackageManager(Toplevel):
    def __init__(self, mainWin, packageNamesWithNewerFileDates):
        super(DialogPackageManager, self).__init__(mainWin.parent)

        self.ENABLE = _("Enable")
        self.DISABLE = _("Disable")
        self.parent = mainWin.parent
        self.cntlr = mainWin
        self.webCache = mainWin.webCache

        # copy plugins for temporary display
        self.packagesConfig = PackageManager.packagesConfig
        self.packagesConfigChanged = False
        self.packageNamesWithNewerFileDates = packageNamesWithNewerFileDates

        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", self.parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))

        self.title(_("Taxonomy Packages Manager"))
        frame = Frame(self)

        # left button frame
        buttonFrame = Frame(frame, width=40)
        buttonFrame.columnconfigure(0, weight=1)
        addLabel = Label(buttonFrame, text=_("Find taxonomy packages:"), wraplength=64, justify="center")
        if not self.webCache.workOffline:
            addSelectFromRegistryButton = Button(buttonFrame, text=_("Select"), command=self.selectFromRegistry)
            ToolTip(addSelectFromRegistryButton, text=_("Select package from the XBRL Package Registry."), wraplength=240)
        addLocalButton = Button(buttonFrame, text=_("Locally"), command=self.findLocally)
        ToolTip(addLocalButton, text=_("File chooser allows selecting taxonomy packages to add (or reload), from the local file system.  "
                                       "Select either a PWD or prior taxonomy package zip file, or a taxonomy manifest (.taxonomyPackage.xml) within an unzipped taxonomy package.  "), wraplength=360)
        addWebButton = Button(buttonFrame, text=_("On Web"), command=self.findOnWeb)
        ToolTip(addWebButton, text=_("Dialog to enter URL full path to load (or reload) package, from the web or local file system.  "
                                     "URL may be either a PWD or prior taxonomy package zip file, or a taxonomy manifest (.taxonomyPackage.xml) within an unzipped taxonomy package.  "), wraplength=360)
        manifestNameButton = Button(buttonFrame, text=_("Manifest"), command=self.manifestName)
        ToolTip(manifestNameButton, text=_("Provide pre-PWD non-standard archive manifest file name pattern (e.g., *taxonomyPackage.xml).  "
                                           "Uses unix file name pattern matching.  "
                                           "Multiple manifest files are supported in pre-PWD archives (such as oasis catalogs).  "
                                           "(Replaces pre-PWD search for either .taxonomyPackage.xml or catalog.xml).  "), wraplength=480)
        self.manifestNamePattern = ""
        addLabel.grid(row=0, column=0, pady=4)
        selBtnRow = 1
        if not self.webCache.workOffline:
            addSelectFromRegistryButton.grid(row=selBtnRow, column=0, pady=4)
            selBtnRow += 1
        addLocalButton.grid(row=selBtnRow, column=0, pady=4)
        selBtnRow += 1
        addWebButton.grid(row=selBtnRow, column=0, pady=4)
        selBtnRow += 1
        manifestNameButton.grid(row=selBtnRow, column=0, pady=4)
        selBtnRow += 1
        buttonFrame.grid(row=0, column=0, rowspan=3, sticky=(N, S, W), padx=3, pady=3)

        # right tree frame (packages already known to arelle)
        packagesFrame = Frame(frame, width=700)
        vScrollbar = Scrollbar(packagesFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(packagesFrame, orient=HORIZONTAL)
        self.packagesView = Treeview(packagesFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set, height=7)
        self.packagesView.grid(row=0, column=0, sticky=(N, S, E, W))
        self.packagesView.bind('<<TreeviewSelect>>', self.packageSelect)
        hScrollbar["command"] = self.packagesView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.packagesView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        packagesFrame.columnconfigure(0, weight=1)
        packagesFrame.rowconfigure(0, weight=1)
        packagesFrame.grid(row=0, column=1, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        self.packagesView.focus_set()

        self.packagesView.column("#0", width=190, anchor="w")
        self.packagesView.heading("#0", text=_("Name"))
        self.packagesView["columns"] = ("ver", "status", "date", "update", "descr")
        self.packagesView.column("ver", width=80, anchor="w", stretch=False)
        self.packagesView.heading("ver", text=_("Version"))
        self.packagesView.column("status", width=50, anchor="w", stretch=False)
        self.packagesView.heading("status", text=_("Status"))
        self.packagesView.column("date", width=170, anchor="w", stretch=False)
        self.packagesView.heading("date", text=_("File Date"))
        self.packagesView.column("update", width=50, anchor="w", stretch=False)
        self.packagesView.heading("update", text=_("Update"))
        self.packagesView.column("descr", width=200, anchor="w", stretch=False)
        self.packagesView.heading("descr", text=_("Description"))

        remappingsFrame = Frame(frame)
        vScrollbar = Scrollbar(remappingsFrame, orient=VERTICAL)
        hScrollbar = Scrollbar(remappingsFrame, orient=HORIZONTAL)
        self.remappingsView = Treeview(remappingsFrame, xscrollcommand=hScrollbar.set, yscrollcommand=vScrollbar.set, height=5)
        self.remappingsView.grid(row=0, column=0, sticky=(N, S, E, W))
        hScrollbar["command"] = self.remappingsView.xview
        hScrollbar.grid(row=1, column=0, sticky=(E,W))
        vScrollbar["command"] = self.remappingsView.yview
        vScrollbar.grid(row=0, column=1, sticky=(N,S))
        remappingsFrame.columnconfigure(0, weight=1)
        remappingsFrame.rowconfigure(0, weight=1)
        remappingsFrame.grid(row=1, column=1, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        self.remappingsView.focus_set()

        self.remappingsView.column("#0", width=200, anchor="w")
        self.remappingsView.heading("#0", text=_("Prefix"))
        self.remappingsView["columns"] = ("remapping")
        self.remappingsView.column("remapping", width=500, anchor="w", stretch=False)
        self.remappingsView.heading("remapping", text=_("Remapping"))

        # bottom frame package info details
        packageInfoFrame = Frame(frame, width=700)
        packageInfoFrame.columnconfigure(1, weight=1)

        self.packageNameLabel = Label(packageInfoFrame, wraplength=600, justify="left",
                                      font=font.Font(family='Helvetica', size=12, weight='bold'))
        self.packageNameLabel.grid(row=0, column=0, columnspan=6, sticky=W)
        self.packageVersionHdr = Label(packageInfoFrame, text=_("version:"), state=DISABLED)
        self.packageVersionHdr.grid(row=1, column=0, sticky=W)
        self.packageVersionLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packageVersionLabel.grid(row=1, column=1, columnspan=5, sticky=W)
        self.packageLicenseHdr = Label(packageInfoFrame, text=_("license:"), state=DISABLED)
        self.packageLicenseHdr.grid(row=2, column=0, sticky=W)
        self.packageLicenseLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packageLicenseLabel.grid(row=2, column=1, columnspan=5, sticky=W)
        self.packageDescrHdr = Label(packageInfoFrame, text=_("description:"), state=DISABLED)
        self.packageDescrHdr.grid(row=3, column=0, sticky=W)
        self.packageDescrLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packageDescrLabel.grid(row=3, column=1, columnspan=5, sticky=W)
        self.packagePrefixesHdr = Label(packageInfoFrame, text=_("prefixes:"), state=DISABLED)
        self.packagePrefixesHdr.grid(row=4, column=0, sticky=W)
        self.packagePrefixesLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packagePrefixesLabel.grid(row=4, column=1, columnspan=5, sticky=W)
        ToolTip(self.packagePrefixesLabel, text=_("List of prefixes that this package remaps."), wraplength=240)
        self.packageUrlHdr = Label(packageInfoFrame, text=_("URL:"), state=DISABLED)
        self.packageUrlHdr.grid(row=5, column=0, sticky=W)
        self.packageUrlLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packageUrlLabel.grid(row=5, column=1, columnspan=5, sticky=W)
        ToolTip(self.packageUrlLabel, text=_("URL of taxonomy package (local file path or web loaded file)."), wraplength=240)
        self.packageDateHdr = Label(packageInfoFrame, text=_("date:"), state=DISABLED)
        self.packageDateHdr.grid(row=6, column=0, sticky=W)
        self.packageDateLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.packageDateLabel.grid(row=6, column=1, columnspan=5, sticky=W)
        ToolTip(self.packageDateLabel, text=_("Filesystem date of currently loaded package file (with parenthetical node when an update is available)."), wraplength=240)
        self.publisherHdr = Label(packageInfoFrame, text=_("publisher:"), state=DISABLED)
        self.publisherHdr.grid(row=7, column=0, sticky=W)
        self.publisherLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.publisherLabel.grid(row=7, column=1, columnspan=5, sticky=W)
        ToolTip(self.publisherLabel, text=_("Publisher of currently loaded package file."), wraplength=240)
        self.publicationDateHdr = Label(packageInfoFrame, text=_("publication date:"), state=DISABLED)
        self.publicationDateHdr.grid(row=8, column=0, sticky=W)
        self.publicationDateLabel = Label(packageInfoFrame, wraplength=600, justify="left")
        self.publicationDateLabel.grid(row=8, column=1, columnspan=5, sticky=W)
        ToolTip(self.publicationDateLabel, text=_("Publication date"), wraplength=240)
        self.packageEnableButton = Button(packageInfoFrame, text=self.ENABLE, state=DISABLED, command=self.packageEnable)
        ToolTip(self.packageEnableButton, text=_("Enable/disable package."), wraplength=240)
        self.packageEnableButton.grid(row=9, column=1, sticky=E)
        self.packageMoveUpButton = Button(packageInfoFrame, text=_("Move Up"), state=DISABLED, command=self.packageMoveUp)
        ToolTip(self.packageMoveUpButton, text=_("Move package up (above other remappings)."), wraplength=240)
        self.packageMoveUpButton.grid(row=9, column=2, sticky=E)
        self.packageMoveDownButton = Button(packageInfoFrame, text=_("Move Down"), state=DISABLED, command=self.packageMoveDown)
        ToolTip(self.packageMoveDownButton, text=_("Move package down (below other remappings)."), wraplength=240)
        self.packageMoveDownButton.grid(row=9, column=3, sticky=E)
        self.packageReloadButton = Button(packageInfoFrame, text=_("Reload"), state=DISABLED, command=self.packageReload)
        ToolTip(self.packageReloadButton, text=_("Reload/update package."), wraplength=240)
        self.packageReloadButton.grid(row=9, column=4, sticky=E)
        self.packageRemoveButton = Button(packageInfoFrame, text=_("Remove"), state=DISABLED, command=self.packageRemove)
        ToolTip(self.packageRemoveButton, text=_("Remove package from packages table (does not erase the package file)."), wraplength=240)
        self.packageRemoveButton.grid(row=9, column=5, sticky=E)
        packageInfoFrame.grid(row=2, column=0, columnspan=5, sticky=(N, S, E, W), padx=3, pady=3)
        packageInfoFrame.config(borderwidth=4, relief="groove")

        okButton = Button(frame, text=_("Close"), command=self.ok)
        ToolTip(okButton, text=_("Accept and changes (if any) and close dialog."), wraplength=240)
        cancelButton = Button(frame, text=_("Cancel"), command=self.close)
        ToolTip(cancelButton, text=_("Cancel changes (if any) and close dialog."), wraplength=240)
        okButton.grid(row=3, column=3, sticky=(S,E), pady=3)
        cancelButton.grid(row=3, column=4, sticky=(S,E), pady=3, padx=3)

        enableDisableFrame = Frame(frame)
        enableDisableFrame.grid(row=3, column=1, sticky=(S,W), pady=3)
        enableAllButton = Button(enableDisableFrame, text=_("Enable All"), command=self.enableAll)
        ToolTip(enableAllButton, text=_("Enable all packages."), wraplength=240)
        disableAllButton = Button(enableDisableFrame, text=_("Disable All"), command=self.disableAll)
        ToolTip(disableAllButton, text=_("Disable all packages."), wraplength=240)
        enableAllButton.grid(row=1, column=1)
        disableAllButton.grid(row=1, column=2)

        self.loadTreeViews()

        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))
        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(0, weight=0)
        frame.columnconfigure(1, weight=1)
        frame.rowconfigure(0, weight=1)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        window.rowconfigure(0, weight=1)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def loadTreeViews(self):
        self.selectedModule = None

        # clear previous treeview entries
        for previousNode in self.packagesView.get_children(""):
            self.packagesView.delete(previousNode)

        for i, packageInfo in enumerate(self.packagesConfig.get("packages", [])):
            name = packageInfo.get("name", "package{}".format(i))
            node = self.packagesView.insert("", "end", "_{}".format(i), text=name)
            self.packagesView.set(node, "ver", packageInfo.get("version"))
            self.packagesView.set(node, "status", packageInfo.get("status"))
            self.packagesView.set(node, "date", packageInfo.get("fileDate"))
            if name in self.packageNamesWithNewerFileDates:
                self.packagesView.set(node, "update", _("available"))
            self.packagesView.set(node, "descr", packageInfo.get("description"))

        # clear previous treeview entries
        for previousNode in self.remappingsView.get_children(""):
            self.remappingsView.delete(previousNode)

        for i, remappingItem in enumerate(sorted(self.packagesConfig.get("remappings", {}).items())):
            prefix, remapping = remappingItem
            node = self.remappingsView.insert("", "end", prefix, text=prefix)
            self.remappingsView.set(node, "remapping", remapping)

        self.packageSelect()  # clear out prior selection

    def ok(self, event=None):
        if self.packagesConfigChanged:
            PackageManager.packagesConfig = self.packagesConfig
            PackageManager.packagesConfigChanged = True
            self.cntlr.onPackageEnablementChanged()
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()

    def packageSelect(self, *args):
        node = (self.packagesView.selection() or (None,))[0]
        try:
            nodeIndex = int(node[1:])
        except (ValueError, TypeError):
            nodeIndex = -1
        if 0 <= nodeIndex < len(self.packagesConfig["packages"]):
            packageInfo = self.packagesConfig["packages"][nodeIndex]
            self.selectedPackageIndex = nodeIndex
            name = packageInfo["name"]
            self.packageNameLabel.config(text=name)
            self.packageVersionHdr.config(state=ACTIVE)
            self.packageVersionLabel.config(text=packageInfo["version"])
            self.packageLicenseHdr.config(state=ACTIVE)
            self.packageLicenseLabel.config(text=packageInfo.get("license"))
            self.packageDescrHdr.config(state=ACTIVE)
            self.packageDescrLabel.config(text=packageInfo["description"])
            self.packagePrefixesHdr.config(state=ACTIVE)
            self.packagePrefixesLabel.config(text=', '.join(packageInfo["remappings"].keys()))
            self.packageUrlHdr.config(state=ACTIVE)
            self.packageUrlLabel.config(text=packageInfo["URL"])
            self.packageDateHdr.config(state=ACTIVE)
            self.packageDateLabel.config(text=packageInfo["fileDate"] + " " +
                    (_("(an update is available)") if name in self.packageNamesWithNewerFileDates else ""))
            self.publisherHdr.config(state=ACTIVE)
            _publisher = ''
            if packageInfo.get("publisher"):
                _publisher += packageInfo["publisher"]
            if packageInfo.get("publisherCountry"):
                _publisher += ", " + packageInfo["publisherCountry"]
            if packageInfo.get("publisherURL"):
                _publisher += ". " + packageInfo["publisherURL"]
            self.publisherLabel.config(text=_publisher)
            self.publicationDateHdr.config(state=ACTIVE)
            self.publicationDateLabel.config(text=packageInfo.get("publicationDate",''))
            self.packageEnableButton.config(state=ACTIVE,
                                           text={"enabled":self.DISABLE,
                                                 "disabled":self.ENABLE}[packageInfo["status"]])
            self.packageMoveUpButton.config(state=ACTIVE if 0 < nodeIndex else DISABLED)
            self.packageMoveDownButton.config(state=ACTIVE if nodeIndex < (len(self.packagesConfig["packages"]) - 1) else DISABLED)
            self.packageReloadButton.config(state=ACTIVE)
            self.packageRemoveButton.config(state=ACTIVE)
        else:
            self.selectedPackageIndex = -1
            self.packageNameLabel.config(text="")
            self.packageVersionHdr.config(state=DISABLED)
            self.packageVersionLabel.config(text="")
            self.packageLicenseHdr.config(state=DISABLED)
            self.packageLicenseLabel.config(text="")
            self.packageDescrHdr.config(state=DISABLED)
            self.packageDescrLabel.config(text="")
            self.packagePrefixesHdr.config(state=DISABLED)
            self.packagePrefixesLabel.config(text="")
            self.packageUrlHdr.config(state=DISABLED)
            self.packageUrlLabel.config(text="")
            self.packageDateHdr.config(state=DISABLED)
            self.packageDateLabel.config(text="")
            self.publisherHdr.config(state=DISABLED)
            self.publisherLabel.config(text="")
            self.publicationDateHdr.config(state=DISABLED)
            self.publicationDateLabel.config(text="")

            self.packageEnableButton.config(state=DISABLED, text=self.ENABLE)
            self.packageMoveUpButton.config(state=DISABLED)
            self.packageMoveDownButton.config(state=DISABLED)
            self.packageReloadButton.config(state=DISABLED)
            self.packageRemoveButton.config(state=DISABLED)


    def selectFromRegistry(self):
        choices = [] # list of tuple of (file name, description)
        uiLang = (self.cntlr.config.get("userInterfaceLangOverride") or self.cntlr.modelManager.defaultLang or "en")[:2]
        def langLabel(labels):
            if not labels:
                return ""
            for _lang in uiLang, "en":
                for label in labels:
                    if label["Language"].startswith(_lang):
                         return label["Label"]
            for label in labels:
                 return label["Label"]
            return ""
        try:
            with open(self.webCache.getfilename(STANDARD_PACKAGES_URL, reload=True), 'r', errors='replace') as fh:
                regPkgs = json.load(fh) # always reload
            for pkgTxmy in regPkgs.get("taxonomies", []):
                _name = langLabel(pkgTxmy["Name"])
                _description = langLabel(pkgTxmy.get("Description"))
                _version = pkgTxmy.get("Version")
                _license = pkgTxmy.get("License",{}).get("Name")
                _url = pkgTxmy.get("Links",{}).get("AuthoritativeURL")
                choices.append((_name,
                                "name: {}\ndescription: {}\nversion: {}\nlicense: {}".format(
                                        _name, _description, _version, _license),
                                _url,  _version, _description, _license))
            self.loadPackageUrl(DialogOpenArchive.selectPackage(self, choices))
        except (TypeError) as err:
            messagebox.showwarning(_("Unable to retrieve standard packages.  "),
                                   _("Standard packages URL is not accessible, please check if online:\n\n{0}.")
                                   .format(STANDARD_PACKAGES_URL),
                                   parent=self)


    def findLocally(self):
        initialdir = self.cntlr.pluginDir # default plugin directory
        #if not self.cntlr.isMac: # can't navigate within app easily, always start in default directory
        initialdir = self.cntlr.config.setdefault("packageOpenDir", initialdir)
        filename = self.cntlr.uiFileDialog("open",
                                           parent=self,
                                           title=_("Choose taxonomy package file"),
                                           initialdir=initialdir,
                                           filetypes=[(_("Taxonomy package files (*.zip)"), "*.zip"),
                                                      (_("PWD Manifest (taxonomyPackage.xml)"), "taxonomyPackage.xml"),
                                                      (_("pre-PWD Manifest (*.taxonomyPackage.xml)"), "*.taxonomyPackage.xml"),
                                                      (_("pre-PWD Oasis Catalog (*catalog.xml)"), "*catalog.xml")],
                                           defaultextension=".zip")
        if filename:
            # check if a package is selected (any file in a directory containing an __init__.py
            self.cntlr.config["packageOpenDir"] = os.path.dirname(filename)
            packageInfo = PackageManager.packageInfo(self.cntlr, filename, packageManifestName=self.manifestNamePattern)
            self.loadFoundPackageInfo(packageInfo, filename)


    def findOnWeb(self):
        self.loadPackageUrl(DialogURL.askURL(self))

    def loadPackageUrl(self, url):
        if url:  # url is the in-cache or local file
            packageInfo = PackageManager.packageInfo(self.cntlr, url, packageManifestName=self.manifestNamePattern)
            self.cntlr.showStatus("") # clear web loading status
            self.loadFoundPackageInfo(packageInfo, url)

    def manifestName(self):
        self.manifestNamePattern = simpledialog.askstring(_("Archive manifest file name pattern"),
                                                          _("Provide non-standard archive manifest file name pattern (e.g., *taxonomyPackage.xml).  \n"
                                                            "Uses unix file name pattern matching.  \n"
                                                            "Multiple manifest files are supported in archive (such as oasis catalogs).  \n"
                                                            "(If blank, search for either .taxonomyPackage.xml or catalog.xml).  "),
                                                          initialvalue=self.manifestNamePattern,
                                                          parent=self)

    def loadFoundPackageInfo(self, packageInfo, url):
        if packageInfo and packageInfo.get("name"):
            self.addPackageInfo(packageInfo)
            self.loadTreeViews()
        else:
            messagebox.showwarning(_("Package is not itself a taxonomy package.  "),
                                   _("File does not itself contain a manifest file: \n\n{0}\n\n  "
                                     "If opening an archive file, the manifest file search pattern currently is \"\", please press \"Manifest\" to change manifest file name pattern, e.g.,, \"*.taxonomyPackage.xml\", if needed.  ")
                                   .format(url),
                                   parent=self)

    def removePackageInfo(self, name, version):
        # find package entry
        packagesList = self.packagesConfig["packages"]
        j = -1
        for i, packageInfo in enumerate(packagesList):
            if packageInfo['name'] == name and packageInfo['version'] == version:
                j = i
                break
        if 0 <= j < len(packagesList):
            del self.packagesConfig["packages"][i]
            self.packagesConfigChanged = True

    def addPackageInfo(self, packageInfo):
        name = packageInfo["name"]
        version = packageInfo["version"]
        self.removePackageInfo(name, version)  # remove any prior entry for this package
        self.packageNamesWithNewerFileDates.discard(name) # no longer has an update available
        self.packagesConfig["packages"].append(packageInfo)
        PackageManager.rebuildRemappings(self.cntlr)
        self.packagesConfigChanged = True

    def packageEnable(self):
        if 0 <= self.selectedPackageIndex < len(self.packagesConfig["packages"]):
            packageInfo = self.packagesConfig["packages"][self.selectedPackageIndex]
            if self.packageEnableButton['text'] == self.ENABLE:
                packageInfo["status"] = "enabled"
                self.packageEnableButton['text'] = self.DISABLE
            elif self.packageEnableButton['text'] == self.DISABLE:
                packageInfo["status"] = "disabled"
                self.packageEnableButton['text'] = self.ENABLE
            self.packagesConfigChanged = True
            PackageManager.rebuildRemappings(self.cntlr)
            self.loadTreeViews()

    def packageMoveUp(self):
        if 1 <= self.selectedPackageIndex < len(self.packagesConfig["packages"]):
            packages = self.packagesConfig["packages"]
            packageInfo = packages[self.selectedPackageIndex]
            del packages[self.selectedPackageIndex]
            packages.insert(self.selectedPackageIndex -1, packageInfo)
            self.packagesConfigChanged = True
            PackageManager.rebuildRemappings(self.cntlr)
            self.loadTreeViews()

    def packageMoveDown(self):
        if 0 <= self.selectedPackageIndex < len(self.packagesConfig["packages"]) - 1:
            packages = self.packagesConfig["packages"]
            packageInfo = packages[self.selectedPackageIndex]
            del packages[self.selectedPackageIndex]
            packages.insert(self.selectedPackageIndex + 1, packageInfo)
            self.packagesConfigChanged = True
            PackageManager.rebuildRemappings(self.cntlr)
            self.loadTreeViews()

    def packageReload(self):
        if 0 <= self.selectedPackageIndex < len(self.packagesConfig["packages"]):
            packageInfo = self.packagesConfig["packages"][self.selectedPackageIndex]
            url = packageInfo.get("URL")
            if url:
                packageInfo = PackageManager.packageInfo(self.cntlr, url, reload=True, packageManifestName=packageInfo.get("manifestName"))
                if packageInfo:
                    self.addPackageInfo(packageInfo)
                    PackageManager.rebuildRemappings(self.cntlr)
                    self.loadTreeViews()
                    self.cntlr.showStatus(_("{0} reloaded").format(packageInfo.get("name")), clearAfter=5000)
                else:
                    messagebox.showwarning(_("Package error"),
                                           _("File or package cannot be reloaded: \n\n{0}")
                                           .format(url),
                                           parent=self)

    def packageRemove(self):
        if 0 <= self.selectedPackageIndex < len(self.packagesConfig["packages"]):
            packageInfo = self.packagesConfig["packages"][self.selectedPackageIndex]
            self.removePackageInfo(packageInfo["name"], packageInfo["version"])
            self.packagesConfigChanged = True
            PackageManager.rebuildRemappings(self.cntlr)
            self.loadTreeViews()

    def enableAll(self):
        self.enableDisableAll(True)

    def disableAll(self):
        self.enableDisableAll(False)

    def enableDisableAll(self, doEnable):
        for iPkg in range(len(self.packagesConfig["packages"])):
            packageInfo = self.packagesConfig["packages"][iPkg]
            if doEnable:
                packageInfo["status"] = "enabled"
                self.packageEnableButton['text'] = self.DISABLE
            else:
                packageInfo["status"] = "disabled"
                self.packageEnableButton['text'] = self.ENABLE
        self.packagesConfigChanged = True
        PackageManager.rebuildRemappings(self.cntlr)
        self.loadTreeViews()
